import dataclasses
from typing import List, Any, Dict, Union

from spinta.components import Page, Property
from spinta.core.ufuncs import Env, Expr
from spinta.types.datatype import DataType, Object
from spinta.utils.schema import NA


class QueryPage:
    page_: Page
    size: int
    select: list
    sort: list

    def __init__(self):
        self.size = 0
        self.select = []
        self.sort = []
        self.page_ = Page()


class QueryParams:
    prioritize_uri: bool = False
    lang_priority: List[str] = None
    expand: Expr = None
    default_expand: bool = False
    lang: List = None
    push: bool = False


class Selected:
    # Item index in select list.
    item: int = None
    # Model property if a property is selected.
    prop: Property = None
    # A value or an Expr for further processing on selected value.
    prep: Any = NA

    def __init__(
        self,
        item: int = None,
        prop: Property = None,
        # `prop` can be Expr or any other value.
        prep: Any = NA,
    ):
        self.item = item
        self.prop = prop
        self.prep = prep

    def __repr__(self):
        return self.debug()

    def __eq__(self, other):
        if isinstance(other, Selected):
            return self.item == other.item and self.prop == other.prop and self.prep == other.prep
        return False

    def debug(self, indent: str = ''):
        prop = self.prop.place if self.prop else 'None'
        if isinstance(self.prep, Selected):
            return (
                f'{indent}Selected('
                f'item={self.item}, '
                f'prop={prop}, '
                f'prep=...)\n'
                   ) + self.prep.debug(indent + '  ')
        elif isinstance(self.prep, (tuple, list)):
            return (
                f'{indent}Selected('
                f'item={self.item}, '
                f'prop={prop}, '
                f'prep={type(self.prep).__name__}...)\n'
            ) + ''.join([
                p.debug(indent + '- ')
                if isinstance(p, Selected)
                else str(p)
                for p in self.prep
            ])
        else:
            return (
                f'{indent}Selected('
                f'item={self.item}, '
                f'prop={prop}, '
                f'prep={self.prep})\n'
            )


class BaseQueryBuilder(Env):
    page: QueryPage
    expand: List[Property] = None
    query_params: QueryParams = None
    # `resolved` is used to map which prop.place properties are already
    # resolved, usually it maps to Selected, but different DataType's can return
    # different results.
    resolved: Dict[str, Selected]
    selected: Dict[str, Selected] = None

    def init_query_params(self, params: QueryParams):
        if params is None:
            params = QueryParams()
        self.query_params = params
        self._set_expanded_properties(params)

    # By default, when expand = None we think that nothing is expanded
    # in case we allow default_expand then we set it to empty list
    # if expand is empty list, we assume all are expanded
    # ?expand() will result in [] and ?expand(name) will result in [name]
    def _set_expanded_properties(self, params: QueryParams):
        prop_expr = params.expand
        if prop_expr is None and not params.default_expand:
            self.expand = None
            return

        self.expand = []
        if prop_expr is not None:
            self.expand = self.resolve(prop_expr)


class Star:
    def __str__(self):
        return '*'


class Func:
    pass


@dataclasses.dataclass
class ReservedProperty(Func):
    dtype: DataType
    param: str


@dataclasses.dataclass
class NestedProperty(Func):
    # Used to mark object nesting
    # for example object -> datatype

    left: Union[Object]
    right: Any


@dataclasses.dataclass
class ResultProperty(Func):
    # Used when result is calculated at ResultBuilder level
    # for example: checksum()
    expr: Expr



from __future__ import annotations

import os
from copy import deepcopy
from typing import Any

from lxml.etree import _Element

from spinta.components import Context
from lxml import etree, objectify
from urllib.request import urlopen

from spinta.core.ufuncs import Expr
from spinta.utils.naming import to_property_name, to_model_name, Deduplicator, to_dataset_name


# mapping of XSD datatypes to DSA datatypes
# XSD datatypes: https://www.w3.org/TR/xmlschema11-2/#built-in-datatypes
# DSA datatypes: https://atviriduomenys.readthedocs.io/dsa/duomenu-tipai.html#duomenu-tipai
# TODO: finish mapping and make sure all things are mapped correctly
DATATYPES_MAPPING = {
    "string": "string",
    "boolean": "boolean",
    "decimal": "number",
    "float": "number",
    "double": "number",

    # duration has to be mapped to integer. In addition to this, we need a prepare function.
    # This prepare function takes in given duration and turns it into integer (timedelta).
    # More about time entities here:
    # https://atviriduomenys.readthedocs.io/dsa/vienetai.html#laiko-vienetai
    # TODO: add prepare function for duration
    #  https://github.com/atviriduomenys/spinta/issues/594
    "duration": "string",

    "dateTime": "datetime",
    "time": "time",
    "date": "date",
    # format: target type, target column, value in the column
    "gYearMonth": "date;enum;M",
    "gYear": "date;enum;Y",
    "gMonthDay": "string",
    "gDay": "string",
    "gMonth": "string",
    "hexBinary": "string",
    "base64Binary": "binary;prepare;base64",
    "anyURI": "uri",
    "QName": "string",
    "NOTATION": "string",
    "normalizedString": "string",
    "token": "string",
    "language": "string",
    "NMTOKEN": "string",
    "NMTOKENS": "string",
    "Name": "string",
    "NCName": "string",
    "ID": "string",
    "IDREF": "string",
    "IDREFS": "string",
    "ENTITY": "string",
    "ENTITIES": "string",
    "integer": "integer",
    "nonPositiveInteger": "integer",
    "negativeInteger": "integer",
    "long": "integer",
    "int": "integer",
    "short": "integer",
    "byte": "integer",
    "nonNegativeInteger": "integer",
    "unsignedLong": "integer",
    "unsignedInt": "integer",
    "unsignedShort": "integer",
    "unsignedByte": "integer",
    "positiveInteger": "integer",
    "yearMonthDuration": "integer",
    "dayTimeDuration": "integer",
    "dateTimeStamp": "datetime",
    "": "string",

}


class XSDModel:
    """
    Class for creating and handling DSA models from XSD files.

    Example of model data:
    {
        "type": "model",
        "name": "",
        "description": "Įvairūs duomenys",
        "properties": {},
        "external": resource_model_external_info,
        "uri": "http://www.w3.org/2000/01/rdf-schema#Resource"
    }
    """

    def __init__(self, xsd: 'XSDReader', node: _Element = None):
        self.deduplicate = Deduplicator()

        self.xsd: 'XSDReader' = xsd
        self.dataset_name: str = xsd.dataset_name
        self.node: _Element = node
        self.type: str = "model"
        self.name: str | None = None
        self.standalone_name: str | None = None
        self.external: dict | None = None
        self.properties: dict | None = None
        self.uri: str | None = None
        self.description: str | None = None

    def get_data(self):
        model_data: dict = {
            "type": "model",
            "name": self.name
        }
        if self.description is not None:
            model_data["description"] = self.description
        if self.properties is not None:
            model_data["properties"] = self.properties
        if self.external is not None:
            model_data["external"] = self.external
        if self.uri is not None:
            model_data["uri"] = self.uri

        return model_data

    def add_external_info(self, external_name: str):
        self.external = {
            "dataset": self.dataset_name,
            "resource": "resource1",
            "name": external_name
        }

    def set_name(self, name: str):
        self.standalone_name = name
        self.name = f"{self.dataset_name}/{name}"

    def _get_property_type(self, node: _Element) -> str:
        if node.get("ref"):
            return "ref"
        property_type: str = node.get("type")
        if not property_type:
            # this is a self defined simple type, so we take it's base as type
            restrictions: list = node.xpath(f'./*[local-name() = "simpleType"]/*[local-name() = "restriction"]')
            if restrictions:
                property_type = restrictions[0].get("base", "")
            else:
                property_type = ""
        # getting rid of the prefix
        if ":" in property_type:
            property_type = property_type.split(":")[1]

        if property_type in self.xsd.custom_types:
            property_type = self.xsd.custom_types.get(property_type).get("base", "")
        if property_type in DATATYPES_MAPPING:
            property_type = DATATYPES_MAPPING[property_type]
        else:
            property_type = "string"

        return property_type

    def _get_enums(self, node: _Element) -> dict[str, dict[str, Any]]:
        enums = {}
        simple_type = node.xpath(f'./*[local-name() = "simpleType"]')
        if simple_type:
            enums = self.xsd.get_enums_from_simple_type(simple_type[0])
        else:
            node_type = node.get("type")
            if node_type and ":" in node_type:
                node_type = node_type.split(":")[1]
            if node_type in self.xsd.custom_types:
                enums = self.xsd.custom_types[node_type]["enums"]

        return enums

    def _node_to_partial_property(self, node: _Element) -> tuple[str, dict[str, str | bool | dict[str, str | dict[str, Any]]]]:
        """Node can be either element or attribute.
        This function only processes things common to attributes and elements"""
        prop = dict()

        prop["description"] = XSDReader.get_description(node)
        property_name = node.get("name")
        prop["external"] = {"name": property_name}
        property_id = to_property_name(property_name)
        prop["type"] = self._get_property_type(node)
        if ";" in prop["type"]:
            prop_type, target, value = prop["type"].split(";")
            prop["type"] = prop_type
            if target == "enum":
                prop[target] = value
            if target == "prepare":
                prop["external"][target] = Expr(value)

        prop["enums"] = self._get_enums(node)

        return self.deduplicate(property_id), prop

    def attributes_to_properties(
        self,
        element: _Element
    ) -> dict[str, dict[str, str | bool | dict[str, str | dict[str, Any]]]]:
        properties = {}
        attributes = element.xpath(f'./*[local-name() = "attribute"]')
        complex_type = element.xpath(f'./*[local-name() = "complexType"]')
        if complex_type:
            properties.update(self.attributes_to_properties(complex_type[0]))
        for attribute in attributes:

            property_id, prop = self._node_to_partial_property(attribute)
            if not prop["type"]:
                prop["type"] = "string"
            # property source
            prop["external"]["name"] = f'@{prop["external"]["name"]}'

            # property required or not. For attributes only.
            use = attribute.get("use")
            if use == "required":
                required = True
            else:
                required = False
            prop["required"] = required
            properties[property_id] = prop

        # TODO: attribute can be a ref to an externally defined attribute also. Not in RC though
        #  https://github.com/atviriduomenys/spinta/issues/605
        return properties

    def simple_element_to_property(
        self,
        element: _Element,
        is_array: bool = False
    ) -> tuple[str, dict[str, str | bool | dict[str, Any]]]:
        """
        simple element is an element which is either
        an inline or simple type element and doesn't have a ref

        Example of a property:

            'properties': {
                'id': {
                    'type': 'integer',
                    'type_args': None,
                    'required': True,
                    'unique': True,
                    'external': {
                        'name': 'ID',
                        'prepare': NA,
                    }
                },
             },
        """

        property_id, prop = self._node_to_partial_property(element)
        if XSDReader.node_is_ref(element):
            ref: str = element.get("ref")
            if ":" in ref:
                ref = ref.split(":")[1]
            prop["external"]["name"] = ref
            property_id = self.deduplicate(to_property_name(ref))
        prop["external"]["name"] = f'{prop["external"]["name"]}/text()'
        if prop.get("type") == "":
            prop["type"] = "string"
        if XSDReader.is_array(element) or is_array:
            property_id += "[]"
        if XSDReader.is_required(element):
            prop["required"] = True
        else:
            prop["required"] = False
        return property_id, prop

    def properties_from_simple_elements(
        self,
        node: _Element,
        from_root: bool = False,
        properties_required: bool = None
    ) -> dict[str, dict[str, str | bool | dict[str, str | dict[str, Any]]]]:

        properties = {}
        elements = node.xpath(f'./*[local-name() = "element"]')
        for element in elements:
            if (self.xsd.node_is_simple_type_or_inline(element) and
                    not XSDReader.node_is_ref(element) and
                    not (self.xsd._node_is_referenced(element) and from_root)):
                property_id, prop = self.simple_element_to_property(element)
                if from_root:
                    prop["required"] = False
                if properties_required is True:
                    prop["required"] = True
                if properties_required is False:
                    prop["required"] = False
                properties[property_id] = prop
        return properties

    def get_text_property(self, property_type = None) -> dict[str, dict[str, str | dict[str, str]]]:
        if property_type is None:
            property_type = "string"
        return {
            self.deduplicate('text'): {
                'type': property_type,
                'external': {
                    'name': 'text()'
                }
            }}


class XSDReader:

    def __init__(self, path, dataset_name: str):
        self._path: str = path
        self.models: list[XSDModel] = []
        self.custom_types: dict = {}
        self._dataset_given_name: str = dataset_name
        self._set_dataset_and_resource_info()
        self.deduplicate: Deduplicator = Deduplicator()

    @staticmethod
    def get_enums_from_simple_type(node: _Element) -> dict[str, dict[str, Any]]:
        enums = {}
        enum_value = {}
        restrictions = node.xpath(f'./*[local-name() = "restriction"]')
        if restrictions:
            # can be enum
            enumerations = restrictions[0].xpath('./*[local-name() = "enumeration"]')
            for enumeration in enumerations:
                enum_item = {
                    "source": enumeration.get("value")
                }
                enum_value.update({enumeration.get("value"): enum_item})
            enums[""] = enum_value

        return enums

    def _extract_custom_types(self, node: _Element):
        """
        format of custom types:
        {
            "type_name": {
                "base": "type_base"
            }
        }
        """
        custom_types_nodes = node.xpath(f'./*[local-name() = "simpleType"]')
        custom_types = {}
        for type_node in custom_types_nodes:
            type_name = type_node.get("name")
            restrictions = type_node.xpath(f'./*[local-name() = "restriction"]')
            property_type_base = restrictions[0].get("base", "")
            property_type_base = property_type_base.split(":")[1]

            # enums
            enums = self.get_enums_from_simple_type(type_node)

            custom_types[type_name] = {
                "base": property_type_base,
                "enums": enums
            }
        self.custom_types = custom_types

    @staticmethod
    def get_description(element: _Element) -> str:
        annotation = element.xpath(f'./*[local-name() = "annotation"]')
        if not annotation:
            annotation = element.xpath(f'./*[local-name() = "simpleType"]/*[local-name() = "annotation"]')
        description = ""
        if annotation:
            documentation = annotation[0].xpath(f'./*[local-name() = "documentation"]')
            for documentation_part in documentation:
                if documentation_part.text is not None:
                    description = f"{description}{documentation_part.text} "
        return description.strip()

    def _node_is_referenced(self, node):
        # if this node is referenced by some other node
        node_name = node.get('name')
        xpath_search_string = f'//*[@ref="{node_name}"]'
        references = self.root.xpath(xpath_search_string)
        if references:
            return True
        return False

    def _get_referenced_node(self, node):
        ref = node.get("ref")
        if ":" in ref:
            ref = ref.split(":")[1]
        xpath_query = f"//*[@name='{ref}']"
        referenced_node = self.root.xpath(xpath_query)[0]
        return referenced_node


    @staticmethod
    def node_is_ref(node: _Element) -> bool:
        if node.get("ref"):
            return True
        return False

    @staticmethod
    def is_array(element: _Element) -> bool:
        max_occurs: str = element.get("maxOccurs", "1")
        return max_occurs == "unbounded" or int(max_occurs) > 1
    
    def _extract_root(self):
        if self._path.startswith("http"):
            document = etree.parse(urlopen(self._path))
            objectify.deannotate(document, cleanup_namespaces=True)
            self.root = document.getroot()
        else:
            with open(self._path) as file:
                text = file.read()
                self.root = etree.fromstring(bytes(text, encoding='utf-8'))

    def _set_dataset_and_resource_info(self):
        self.dataset_name = to_dataset_name(os.path.splitext(os.path.basename(self._path))[0])
        self.dataset_and_resource_info = {
            'type': 'dataset',
            'name': self.dataset_name,
            'resources': {
                "resource1": {
                    'type': 'xml',
                },
            },
            'given_name': self._dataset_given_name
        }

    def _get_separate_complex_type_node_by_type(self, node_type: str) -> _Element:
        if node_type is not None:
            node_type = node_type.split(":")
            if len(node_type) > 1:
                node_type = node_type[1]
            else:
                node_type = node_type[0]
        if node_type not in DATATYPES_MAPPING:
            complex_types = self.root.xpath(f'./*[local-name() = "complexType"]')
            for node in complex_types:
                if node.get("name") == node_type:
                    return node
    def _get_separate_complex_type_node(self, node: _Element) -> _Element:
        node_type: str | list = node.get('type')
        return self._get_separate_complex_type_node_by_type(node_type)

    def _node_has_separate_complex_type(self, node: _Element) -> bool:
        node_type: str | list = node.get('type')
        if node_type is not None:
            node_type = node_type.split(":")
            if len(node_type) > 1:
                node_type = node_type[1]
            else:
                node_type = node_type[0]
            if node_type not in DATATYPES_MAPPING:
                complex_types = self.root.xpath(f'./*[local-name() = "complexType"]')
                for node in complex_types:
                    if node.get("name") == node_type:
                        return True
        return False

    def node_is_simple_type_or_inline(self, node: _Element) -> bool:
        if self._node_has_separate_complex_type(node):
            return False
        return bool(
            (node.xpath(f'./*[local-name() = "annotation"]') and len(node.getchildren()) == 1) or
            (node.xpath(f'./*[local-name() = "simpleType"]')) or
            (len(node.getchildren()) == 0)
        )

    @staticmethod
    def _is_element(node: _Element) -> bool:
        if node.xpath('local-name()') == "element":
            return True
        return False

    @staticmethod
    def is_required(element: _Element) -> bool:
        min_occurs = int(element.get("minOccurs", 1))
        if min_occurs > 0:
            nillable = element.get("nillable", False)
            if nillable == "true":
                return False
            return True
        return False

    def _properties_from_references(
        self,
        node: _Element,
        model: XSDModel,
        source_path: str = ""
    ) -> dict[str, dict[str, str | bool | dict[str, str | dict[str, Any]]]]:

        properties = {}
        # if len(node) == 1:
        #     # if this model has only one property, which is a reference, we don't create it, but pass it on.
        #     ref_element = node.xpath("./*[@ref]")[0]
        #     ref = ref_element.get("ref")
        #     if ":" in ref:
        #         ref = ref.split(":")[1]
        #     xpath_query = f"//*[@name='{ref}']"
        #     referenced_element = self.root.xpath(xpath_query)[0]
        #     return self._properties_from_references(model=model, source_path=node.get("name"))
        for ref_element in node.xpath("./*[@ref]"):
            referenced_element = self._get_referenced_node(ref_element)

            if self.node_is_simple_type_or_inline(referenced_element):
                property_id, prop = model.simple_element_to_property(referenced_element)
                if not XSDReader.is_required(ref_element):
                    prop["required"] = False
                properties[property_id] = prop
            else:
                is_array = False
                # try:
                    # TODO fix this because it probably doesn't cover all cases, only something like <complexType><sequence><item>
                    #  https://github.com/atviriduomenys/spinta/issues/613
                complex_type = referenced_element.xpath("./*[local-name() = 'complexType']")[0]
                sequences = complex_type.xpath("./*[local-name() = 'sequence']")
                if sequences:
                    sequence = sequences[0]
                else:
                    sequence = None
                if sequence is not None and len(sequence) == 1 and self.node_is_ref(sequence[0]):
                    is_array = XSDReader.is_array(referenced_element)
                    if not is_array:
                        is_array = XSDReader.is_array(complex_type[0][0])
                    new_referenced_element = self._get_referenced_node(complex_type[0][0])
                    referenced_element = new_referenced_element
                    if ref_element.get("name") is not None:
                        source_path += f'/{ref_element.get("name")}'
                # except (TypeError, IndexError):
                #     pass

                if not (XSDReader.is_array(ref_element) or is_array):
                    referenced_model_names = self._create_model(referenced_element, source_path)
                    property_type = "ref"
                else:
                    referenced_element_properties = {
                        to_property_name(model.standalone_name):
                        {
                            "type": "ref",
                            "model": f"{model.name}"
                        }
                    }
                    property_type = "backref"
                    referenced_model_names = self._create_model(referenced_element, source_path,
                                                                additional_properties=referenced_element_properties)

                for referenced_model_name in referenced_model_names:
                    property_id, prop = model.simple_element_to_property(ref_element, is_array=is_array)

                    prop["external"]["name"] = ""
                    prop["type"] = property_type
                    prop["model"] = f"{referenced_model_name}"
                    properties[property_id] = prop

        return properties

    def _split_choice(
        self,
        node: _Element,
        source_path: str,
        additional_properties: dict[str, dict[str, str | bool | dict[str, str]]]
    ) -> list[str]:
        """
        If there are choices in the element,
        we need to split it and create a separate model per each choice
        """

        model_names = []
        node_copy = deepcopy(node)
        if self._node_has_separate_complex_type(node_copy):
            complex_type_node = self._get_separate_complex_type_node(node_copy)
        else:
            complex_type_node = node_copy.xpath(f'./*[local-name() = "complexType"]')[0]

        choice_nodes = complex_type_node.xpath(f'./*[local-name() = "choice"]')

        # if maxOccurs=unbound and there's a second `choice` inside, we have to split that one
        if choice_nodes and choice_nodes[0].get("maxOccurs") == "unbounded":
            choice_nodes = choice_nodes[0].xpath(f'./*[local-name() = "choice"]')
        if choice_nodes:
            choice_node = choice_nodes[0]
        else:
            choice_node = complex_type_node.xpath(f'./*[local-name() = "sequence"]/*[local-name() = "choice"]')[0]
        if len(choice_node) > 0:
            choice_node_copy = deepcopy(choice_node)
            choice_node_parent = choice_node.getparent()
            choice_node_parent.remove(choice_node)

            for choice in choice_node_copy:
                if complex_type_node.xpath(f'./*[local-name() = "sequence"]') and choice_node_copy.xpath(
                        f'./*[local-name() = "sequence"]'):
                    choice_copy = deepcopy(choice)
                    for node_in_choice in choice:
                        choice_node_parent.insert(0, node_in_choice)
                    model_names.extend(self._create_model(node_copy, source_path, additional_properties))
                    for node_in_choice in choice_copy:
                        node_in_choice = choice_node_parent.xpath(f"./*[@name=\'{node_in_choice.get('name')}\']")[0]
                        choice_node_parent.remove(node_in_choice)
                else:
                    choice_node_parent.insert(0, choice)
                    model_names.extend(self._create_model(node_copy, source_path, additional_properties))
                    choice_node_parent.remove(choice)
        return model_names

    def _create_model(
        self,
        node: _Element,
        source_path: str = "",
        additional_properties: dict[str, str | bool | dict[str, str | dict[str, Any]]] = None
    ) -> list[str]:
        """
        Parses an element and makes a model out of it. If it is a complete model, it will be added to the models list.
        """
        model = XSDModel(self)

        if additional_properties is None:
            additional_properties = {}

        properties = {}
        properties.update(additional_properties)

        new_source_path = f"{source_path}/{node.get('name')}"

        model.set_name(self.deduplicate(to_model_name(node.get("name"))))

        # if this is complexType node which has complexContent, with a separate
        # node, we need to join the contents of them both

        description = self.get_description(node)
        properties.update(model.attributes_to_properties(node))

        model_names = []

        if node.xpath(f'./*[local-name() = "complexType"]') or self._node_has_separate_complex_type(node):

            if self._node_has_separate_complex_type(node):
                complex_type_node = self._get_separate_complex_type_node(node)
            else:
                complex_type_node = node.xpath(f'./*[local-name() = "complexType"]')[0]

            # if there is choices, we need to create a separate model for each choice
            choices = complex_type_node.xpath(f'./*[local-name() = "choice"]')
            # if choices is unbounded, we treat it like sequence
            if not choices or choices[0].get("maxOccurs") == "unbounded":
                # if it's a `choice` node with `unbounded`, we treat it the same as sequence node
                if choices:
                    choices = complex_type_node.xpath(f'./*[local-name() = "choice"]/*[local-name() = "choice"]')
                else:
                    choices = complex_type_node.xpath(f'./*[local-name() = "sequence"]/*[local-name() = "choice"]')
            if choices:
                if choices[0].get("maxOccurs") != "unbounded":
                    return self._split_choice(node, source_path, additional_properties=additional_properties)

            # if complextype node's property mixed is true, it allows text inside
            if complex_type_node.get("mixed") == "true":
                properties.update(model.get_text_property())
            if complex_type_node.xpath(f'./*[local-name() = "complexContent"]'):
                # TODO: this is only for the nodes where complex content extension base is abstract.
                #  it's the case for the RC documents, but might be different for other data providers
                #  https://github.com/atviriduomenys/spinta/issues/604

                complex_type_node = complex_type_node.xpath(f'./*[local-name() = "complexContent"]/*[local-name() = "extension"]')[0]
                complex_content_base_name = complex_type_node.get("base")
                complex_content_base_node = self._get_separate_complex_type_node_by_type(complex_content_base_name)
                if complex_content_base_node.xpath(f'./*[local-name() = "sequence"]'):
                    sequence_node = complex_content_base_node.xpath(f'./*[local-name() = "sequence"]')[0]
                    properties.update(model.properties_from_simple_elements(sequence_node))
                # TODO: in this case, it might be something else, not sequence too

            if complex_type_node.xpath(f'./*[local-name() = "sequence"]') \
                    or complex_type_node.xpath(f'./*[local-name() = "all"]')\
                    or complex_type_node.xpath(f'./*[local-name() = "simpleContent"]')\
                    or len(complex_type_node) > 0:
                """
                source: https://stackoverflow.com/questions/36286056/the-difference-between-all-sequence-choice-and-group-in-xsd
                    When to use xsd:all, xsd:sequence, xsd:choice, or xsd:group:

                    Use xsd:all when all child elements must be present, independent of order.
                    Use xsd:sequence when child elements must be present per their occurrence constraints and order does matters.
                    Use xsd:choice when one of the child element must be present.
                    Use xsd:group as a way to wrap any of the above in order to name and reuse in multiple locations within an XSD.
                    Note that occurrence constraints can appear on xsd:all, xsd:sequence, or xsd:choice in addition to the 
                    child elements to achieve various cardinality effects. For example, if minOccurs="0" were added to xsd:element children of xsd:all, element order would be insignificant,
                    but not all child elements would have to be present.
                """
                properties_required = None
                if complex_type_node.xpath(f'./*[local-name() = "sequence"]'):
                    sequence_or_all_nodes = complex_type_node.xpath(f'./*[local-name() = "sequence"]')
                elif complex_type_node.xpath(f'./*[local-name() = "all"]'):
                    sequence_or_all_nodes = complex_type_node.xpath(f'./*[local-name() = "all"]')
                elif complex_type_node.xpath(f'./*[local-name() = "choice"]'):
                    sequence_or_all_nodes = complex_type_node.xpath(f'./*[local-name() = "choice"]')
                    properties_required = False
                elif complex_type_node.xpath(f'./*[local-name() = "simpleContent"]'):
                    sequence_or_all_nodes = complex_type_node.xpath(f'./*[local-name() = "simpleContent"]/*[local-name() = "extension"]')

                    text_property_type = sequence_or_all_nodes[0].get("base")
                    if text_property_type is not None:
                        if ":" in text_property_type:
                            text_property_type = text_property_type.split(":")[1]
                        text_property_type = DATATYPES_MAPPING[text_property_type]
                    text_property = model.get_text_property(text_property_type)
                    properties.update(text_property)
                    # additionally extract properties from attributes, if it's let's say, simpleContent node
                    properties.update(model.attributes_to_properties(sequence_or_all_nodes[0]))
                else:
                    sequence_or_all_nodes = None
                if sequence_or_all_nodes:
                    sequence_or_all_node = sequence_or_all_nodes[0]
                else:
                    sequence_or_all_node = complex_type_node
                sequence_or_all_node_length = len(sequence_or_all_node)
                # There is only one element in the complex node sequence, and it doesn't have annotation.
                # Then we just go deeper and add this model to the next model's path.
                if sequence_or_all_node_length == 1 and not properties:

                    if sequence_or_all_node.xpath(f'./*[local-name() = "element"]'):
                        if not sequence_or_all_node.xpath(f'./*[local-name() = "element"]')[0].get("ref"):
                            element = sequence_or_all_node.xpath(f'./*[local-name() = "element"]')[0]
                            if self.node_is_simple_type_or_inline(element) and not self.node_is_ref(element):
                                properties.update(model.properties_from_simple_elements(sequence_or_all_node, properties_required=properties_required))
                            # check for recursion
                            # TODO: maybe move this to a separate function
                            # TODO: recursion not fully working
                            #  https://github.com/atviriduomenys/spinta/issues/602
                            else:
                                paths = new_source_path.split("/")
                                if not element.get("name") in paths:

                                    # this can sometimes happen when choice node has been split or maybe in some other cases too
                                    return self._create_model(element, source_path=new_source_path)
                                else:
                                    for index, path in enumerate(paths):
                                        if path == element.get("name"):
                                            paths[index] = f"/{path}"
                                    new_source_path = "/".join(paths)

                        else:
                            # TODO: if reference is to an inline or simpleType element,
                            #  and maxOccurs of it is 1,
                            #  then do not create reference, but add to the same

                            # properties.update(
                            #     self._properties_from_references(sequence_or_all_node, model, new_source_path))
                            element = sequence_or_all_node.xpath(f'./*[local-name() = "element"]')[0]
                            element = self._get_referenced_node(element)
                            return self._create_model(element, source_path=new_source_path, additional_properties=additional_properties)

                elif sequence_or_all_node_length > 1 or properties:
                    # properties from simple type or inline elements without references
                    # properties are required for choice where maxOccurs=unbound and maybe some other cases
                    properties.update(model.properties_from_simple_elements(
                        sequence_or_all_node,
                        properties_required=properties_required))

                    # references
                    properties.update(
                        self._properties_from_references(sequence_or_all_node, model, new_source_path))

                    # complex type child nodes - to models
                    for child_node in sequence_or_all_node:
                        if child_node.xpath(f'./*[local-name() = "complexType"]') \
                                or self._node_has_separate_complex_type(child_node):
                            # check for recursion
                            # TODO: maybe move this to a separate function
                            paths = new_source_path.split("/")
                            if not child_node.get("name") in paths:
                                self._create_model(child_node, source_path=new_source_path)
                            else:
                                for index, path in enumerate(paths):
                                    if path == child_node.get("name"):
                                        paths[index] = f"/{path}"
                                new_source_path = "/".join(paths)

        if properties:
            model.properties = properties
            model.add_external_info(external_name=new_source_path)
            model.description = description
            self.models.append(model)

            return [model.name, ]
        return []

    def _add_resource_model(self):
        resource_model = XSDModel(self)
        resource_model.add_external_info(external_name="/")
        resource_model.type = "model"
        resource_model.description = "Įvairūs duomenys"
        resource_model.uri = "http://www.w3.org/2000/01/rdf-schema#Resource"
        resource_model.properties = resource_model.properties_from_simple_elements(self.root, from_root=True)
        if resource_model.properties:
            resource_model.set_name(self.deduplicate(f"Resource"))
            self.models.append(resource_model)

    def _parse_root_node(self):
        # get properties from elements
        # Resource model - special case

        for node in self.root:
            if (
                    self._is_element(node) and
                    (not self.node_is_simple_type_or_inline(node) or self.node_is_ref(node)) and
                    not self._node_is_referenced(node)
            ):
                self._create_model(node)

    def start(self):
        self._extract_root()
        self._extract_custom_types(self.root)
        self._add_resource_model()

        self._parse_root_node()


def read_schema(
    context: Context,
    path: str,
    prepare: str = None,
    dataset_name: str = ''
) -> dict[Any, dict[str, str | dict[str, str | bool | dict[str, str | dict[str, Any]]]]]:
    """
    This reads XSD schema from the url provided in path and yields asd schema models

    For now this is adjusted for XSD schemas of Registrų centras

    Elements can be:
    1. Simple type inline
    2. Simple type
    3. Complex type
    4. Complex type described separately

    Elements can have annotations, which we add as description to either a model or a property.

    1. Simple type.
        a) not a reference. It's a property
        b) a reference
            i. maxOccurs = 1
                A. both referencing and referenced elements has annotations. A ref property
                B. Only one or none of the elements has annotations. A part of the path
            ii. maxOccurs > 1 - a ref property
    2. Inline type.

        a) inline type or a custom type referencing to simple type. A property
        b) complex type defined separately of this element.

        If it's a root element, it's then added as a property to a special Resource model.

        Simple type can have annotation too.

    3. Complex type
        If element is complex type, it can be either a property or a model,
        or a part of a source path for a property or model, depending on other factors

            a) element has attributes. Then it's a model and attributes are properties.
            b) element has a sequence
                i. There is more than one element in the sequence. Then it's a model/end part of the model.
                   We treat child elements in the same way as we do root elements (recursion) and build a path
                ii. There is only one element in the sequence. It can then be a property, a model or an intermediate
                    in the path of a model.
                    Options:
                        A. That element isn't a ref.
                            We treat child elements in the same way as we do root elements (recursion) and build a path
                        B. That element is a ref.
                            B1. If maxOccurs > 1, this is a model, and that other element is also a model
                            B2. If maxOccurs = 1
                                B21. If both have annotation, then it's a model
                                B22. If only one of them has annotation, it's a part of a path
                            todo finish defining behaviours of different options for sequences

            c) element has a choice. todo define behaviour here

    4. complex type described separately

    We will build a list of models
    (another option was to make a generator for parsing models, and going deeper, but that would
    be more complex when returning models. although this option is also possible, but it can
    be reworked into this from an option with a list)

    If some model has references in the sequence, we need to also add those as models,
    mark their type as backref. In this case, we add them at the moment we meet them
    as xsd refs, because this way we will know that they are backrefs. We also need to add refs on
    those models to the current model.

    Element can be as a ref in more than one other element.

    Other things to handle: Resource model, custom types, enumerations, choices

    Element can be turned into:
        1. A property (including reference properties)
        2. A model (including referenced models)
        3. A part of another model (as a path)
        4. A part of another property (as a path)

    Attribute can only be turned into a property

    A property can also be text()
    """
    xsd = XSDReader(path, dataset_name)

    xsd.start()

    yield None, xsd.dataset_and_resource_info

    for parsed_model in xsd.models:

        yield None, parsed_model.get_data()



from abc import ABC, abstractmethod
from sqlalchemy.dialects.postgresql import UUID
from typing import List, Tuple

import sqlalchemy as sa
from alembic.operations import Operations


class MigrationAction(ABC):
    @abstractmethod
    def execute(self, op: Operations):
        pass


class CreateTableMigrationAction(MigrationAction):
    def __init__(self, table_name: str, columns: List[sa.Column]):
        self.table_name = table_name
        self.columns = columns

    def execute(self, op: Operations):
        op.create_table(self.table_name, *self.columns)


class DropTableMigrationAction(MigrationAction):
    def __init__(self, table_name: str):
        self.table_name = table_name

    def execute(self, op: Operations):
        op.drop_table(table_name=self.table_name)


class RenameTableMigrationAction(MigrationAction):
    def __init__(self, old_table_name: str, new_table_name: str):
        self.old_table_name = old_table_name
        self.new_table_name = new_table_name

    def execute(self, op: Operations):
        op.rename_table(old_table_name=self.old_table_name, new_table_name=self.new_table_name)


class AddColumnMigrationAction(MigrationAction):
    def __init__(self, table_name: str, column: sa.Column):
        self.table_name = table_name
        self.column = column

    def execute(self, op: Operations):
        op.add_column(table_name=self.table_name, column=self.column)


class DropColumnMigrationAction(MigrationAction):
    def __init__(self, table_name: str, column_name: str):
        self.table_name = table_name
        self.column_name = column_name

    def execute(self, op: Operations):
        op.drop_column(table_name=self.table_name, column_name=self.column_name)


class AlterColumnMigrationAction(MigrationAction):
    def __init__(self, table_name: str, column_name: str, nullable: bool = None, new_column_name: str = None,
                 type_=None, using: str = None):
        self.table_name = table_name
        self.column_name = column_name
        self.nullable = nullable
        self.new_column_name = new_column_name
        self.type_ = type_
        self.using = using

    def execute(self, op: Operations):
        op.alter_column(
            table_name=self.table_name,
            column_name=self.column_name,
            nullable=self.nullable,
            new_column_name=self.new_column_name,
            type_=self.type_,
            postgresql_using=self.using
        )


class DropConstraintMigrationAction(MigrationAction):
    def __init__(self, table_name: str, constraint_name: str):
        self.table_name = table_name
        self.constraint_name = constraint_name

    def execute(self, op: Operations):
        op.drop_constraint(constraint_name=self.constraint_name, table_name=self.table_name)


class DropIndexMigrationAction(MigrationAction):
    def __init__(self, table_name: str, index_name: str):
        self.table_name = table_name
        self.index_name = index_name

    def execute(self, op: Operations):
        op.drop_index(index_name=self.index_name, table_name=self.table_name)


class CreateUniqueConstraintMigrationAction(MigrationAction):
    def __init__(self, table_name: str, constraint_name: str, columns: List[str]):
        self.table_name = table_name
        self.constraint_name = constraint_name
        self.columns = columns

    def execute(self, op: Operations):
        op.create_unique_constraint(constraint_name=self.constraint_name, table_name=self.table_name,
                                    columns=self.columns)


class CreatePrimaryKeyMigrationAction(MigrationAction):
    def __init__(self, table_name: str, constraint_name: str, columns: List[str]):
        self.table_name = table_name
        self.constraint_name = constraint_name
        self.columns = columns

    def execute(self, op: Operations):
        op.create_primary_key(constraint_name=self.constraint_name, table_name=self.table_name, columns=self.columns)


class CreateIndexMigrationAction(MigrationAction):
    def __init__(self, table_name: str, index_name: str, columns: List[str], using: str = None):
        self.table_name = table_name
        self.index_name = index_name
        self.columns = columns
        self.using = using

    def execute(self, op: Operations):
        op.create_index(index_name=self.index_name, table_name=self.table_name, columns=self.columns,
                        postgresql_using=self.using)


class DowngradeTransferDataMigrationAction(MigrationAction):
    def __init__(
        self,
        table_name: str,
        foreign_table_name: str,
        source: sa.Column,
        columns: dict,
        target: str = '_id'
    ):
        target_table = sa.Table(
            table_name,
            sa.MetaData(),
            source._copy(),
            *[sa.Column(key, column.type) for key, column in columns.items()]
        )
        foreign_table = sa.Table(
            foreign_table_name,
            sa.MetaData(),
            sa.Column('_id', UUID),
            *[column._copy() for column in columns.values()]
        )
        self.query = target_table.update().values(
            **{
                key: foreign_table.columns[column.name]
                for key, column in columns.items()
            }
        ).where(
            target_table.columns[source.name] == foreign_table.columns[target]
        )

    def execute(self, op: Operations):
        op.execute(self.query)


class UpgradeTransferDataMigrationAction(MigrationAction):
    def __init__(
        self,
        table_name: str,
        foreign_table_name: str,
        target: sa.Column,
        columns: dict
    ):
        target_table = sa.Table(
            table_name,
            sa.MetaData(),
            target._copy(),
            *[column._copy() for column in columns.values()]
        )
        foreign_table = sa.Table(
            foreign_table_name,
            sa.MetaData(),
            sa.Column('_id', UUID),
            *[sa.Column(key, column.type) for key, column in columns.items()]
        )
        self.query = target_table.update().values(
            **{
                target.name: foreign_table.columns['_id']
            }
        ).where(
            sa.and_(
                *[target_table.columns[column.name] == foreign_table.columns[key] for key, column in columns.items()]
            )
        )

    def execute(self, op: Operations):
        op.execute(self.query)


class CreateForeignKeyMigrationAction(MigrationAction):
    def __init__(self, source_table: str, referent_table: str, constraint_name: str, local_cols: list,
                 remote_cols: list):
        self.constraint_name = constraint_name
        self.source_table = source_table
        self.referent_table = referent_table
        self.local_cols = local_cols
        self.remote_cols = remote_cols

    def execute(self, op: Operations):
        op.create_foreign_key(
            constraint_name=self.constraint_name,
            source_table=self.source_table,
            referent_table=self.referent_table,
            local_cols=self.local_cols,
            remote_cols=self.remote_cols
        )


class RenameSequenceMigrationAction(MigrationAction):
    def __init__(self, old_name: str, new_name: str):
        self.query = f'ALTER SEQUENCE "{old_name}" RENAME TO "{new_name}"'

    def execute(self, op: Operations):
        replaced = self.query.replace(":", "\\:")
        op.execute(replaced)


class TransferJSONDataMigrationAction(MigrationAction):
    def __init__(self, table: str, source: sa.Column, columns: List[Tuple[str, sa.Column]]):
        # Hack to transfer data if columns do not exist
        # Create new table with just required columns exist
        # SQLAlchemy does not allow the use table actions when those columns do not exist
        meta = sa.MetaData()
        temp_table = sa.Table(
            table,
            meta,
            source._copy(),
            *[column._copy() for _, column in columns]
        )
        self.query = temp_table.update().values(
            **{
                temp_table.columns[column.name].name: source[key].astext
                for key, column in columns
            }
        )

    def execute(self, op: Operations):
        op.execute(self.query)


class TransferColumnDataToJSONMigrationAction(MigrationAction):
    def __init__(self, table: str, source: sa.Column, columns: List[Tuple[str, sa.Column]]):
        # # Hack to transfer data if columns do not exist
        # # Create new table with just required columns exist
        # # SQLAlchemy does not allow the use table actions when those columns do not exist
        temp_table = sa.Table(
            table,
            sa.MetaData(),
            source._copy(),
            *[column._copy() for _, column in columns]
        )
        copied_source = temp_table.columns[source.name]
        results = []
        for key, value in columns:
            results.append(key)
            results.append(temp_table.columns[value.name])
        self.query = temp_table.update().values(
            **{
                copied_source.name: copied_source + sa.func.jsonb_build_object(
                    *results
                )
            }
        )

    def execute(self, op: Operations):
        op.execute(self.query)


class AddEmptyAttributeToJSONMigrationAction(MigrationAction):
    def __init__(self, table: str, source: sa.Column, key: str):
        temp_table = sa.Table(
            table,
            sa.MetaData(),
            source._copy(),
        )
        copied_source = temp_table.columns[source.name]
        self.query = temp_table.update().values(
            **{
                copied_source.name: copied_source + sa.func.jsonb_build_object(
                    key, None
                )
            }
        )

    def execute(self, op: Operations):
        op.execute(self.query)


class RenameJSONAttributeMigrationAction(MigrationAction):
    def __init__(self, table: str, source: sa.Column, old_key: str, new_key: str):
        temp_table = sa.Table(
            table,
            sa.MetaData(),
            source._copy(),
        )
        copied_source = temp_table.columns[source.name]
        self.query = temp_table.update().values(
            **{
                copied_source.name: copied_source - old_key + sa.func.jsonb_build_object(new_key, copied_source[old_key])
            }
        ).where(copied_source.has_key(old_key))

    def execute(self, op: Operations):
        op.execute(self.query)


class RemoveJSONAttributeMigrationAction(MigrationAction):
    def __init__(self, table: str, source: sa.Column, key: str):
        temp_table = sa.Table(
            table,
            sa.MetaData(),
            source._copy(),
        )
        copied_source = temp_table.columns[source.name]
        self.query = temp_table.update().values(
            **{
                copied_source.name: copied_source - key
            }
        ).where(copied_source.has_key(key))

    def execute(self, op: Operations):
        op.execute(self.query)


class MigrationHandler:
    def __init__(self):
        self.migrations: List[MigrationAction] = []
        self.foreign_key_migration: List[MigrationAction] = []

    def add_action(self, action: MigrationAction, foreign_key: bool = False):
        if foreign_key:
            self.foreign_key_migration.append(action)
        else:
            self.migrations.append(action)

    def has_constraint_been_dropped(self, constraint_name: str):
        for migration in self.migrations:
            if isinstance(migration, DropConstraintMigrationAction):
                if migration.constraint_name == constraint_name:
                    return True
        for migration in self.foreign_key_migration:
            if isinstance(migration, DropConstraintMigrationAction):
                if migration.constraint_name == constraint_name:
                    return True
        return False

    def run_migrations(self, op: Operations):
        for migration in self.migrations:
            migration.execute(op)
        for migration in self.foreign_key_migration:
            migration.execute(op)



from typing import Union

from spinta.components import Property, Page, UrlParams
from spinta.core.ufuncs import Expr, asttoexpr
from spinta.datasets.components import ExternalBackend
from spinta.types.text.components import Text
from spinta.types.text.helpers import determine_language_property_for_text
from spinta.ufuncs.basequerybuilder.components import BaseQueryBuilder, QueryParams, QueryPage
from spinta.ufuncs.helpers import merge_formulas


def is_expandable_not_expanded(env: BaseQueryBuilder, prop: Property):
    return prop.dtype.expandable and (env.expand is None or (env.expand and prop not in env.expand))


def get_language_column(env: BaseQueryBuilder, dtype: Text):
    default_langs = env.context.get('config').languages
    prop = determine_language_property_for_text(dtype, env.query_params.lang_priority, default_langs)
    column = env.backend.get_column(env.table, prop)
    return column


def get_column_with_extra(env: BaseQueryBuilder, prop: Property):
    if isinstance(prop.dtype, Text):
        return get_language_column(env, prop.dtype)
    return env.backend.get_column(env.table, prop)


def get_page_values(env: BaseQueryBuilder, row: dict):
    if not env.page.page_.filter_only:
        if isinstance(env.model.backend, ExternalBackend):
            return [row[item.prop.external.name] for item in env.page.page_.by.values()]
        else:
            return [row[item.prop.name] for item in env.page.page_.by.values()]


def merge_with_page_selected_list(select_list: list, page: QueryPage):
    merged_selected = select_list or []
    if page.page_.is_enabled and page.select is not None:
        for select in page.select:
            if select not in merged_selected:
                merged_selected.append(select)
    return merged_selected


def merge_with_page_sort(sort: list, page: QueryPage):
    merged_sort = sort or []
    if page.page_.is_enabled and page.sort:
        merged_sort = page.sort
    return merged_sort


def merge_with_page_limit(limit: int, page: QueryPage):
    if page and page.size and page.page_.is_enabled:
        if limit and limit < page.size:
            return limit
        return page.size
    return limit


def update_query_with_url_params(query_params: QueryParams, url_params: UrlParams):
    query_params.prioritize_uri = url_params.fmt.prioritize_uri
    query_params.lang_priority = url_params.accept_langs
    query_params.lang = url_params.lang
    query_params.expand = url_params.expand


def filter_page_values(page: Page):
    new_page = Page()
    for by, page_by in page.by.items():
        # new_page.by[by] = page_by
        if page_by.value is not None:
            new_page.by[by] = page_by
    return new_page


def _create_or_condition(condition_info: list):
    action = condition_info[0]
    page_by = condition_info[1]
    or_null = condition_info[2]

    result = {
        'name': action,
        'args': [{
            'name': 'bind',
            'args': [page_by.prop.name]
        }, page_by.value]
    }

    if or_null != -1:
        result = {
            'name': 'or',
            'args': [
                result,
                {
                    'name': 'eq' if or_null == 1 else 'ne',
                    'args': [{
                        'name': 'bind',
                        'args': [page_by.prop.name]
                    }, None]
                }
            ]}
    return result


def _get_null_action(by: str, needs_null: bool = False):
    if not needs_null:
        return -1

    if by.startswith('-'):
        return 0
    return 1


def get_pagination_compare_query(
    model_page: Page,
) -> Union[Expr, None]:
    filtered = model_page
    item_count = len(filtered.by.keys())
    where_list = []
    for i in range(item_count):
        where_list.append([])
    if not filtered.all_none() or not filtered.first_time:
        for i, (by, page_by) in enumerate(filtered.by.items()):
            for n in range(item_count):
                if n >= i:
                    if n == i:
                        if page_by.value is not None:
                            if by.startswith('-'):
                                where_list[n].append(('lt', page_by, _get_null_action(by)))
                            else:
                                where_list[n].append(('gt', page_by, _get_null_action(by, True)))
                        else:
                            if by.startswith('-'):
                                where_list[n].append(('ne', page_by, _get_null_action(by)))
                    else:
                        where_list[n].append(('eq', page_by, _get_null_action(by)))
    remove_list = []

    for i, (by, value) in enumerate(filtered.by.items()):
        if value.value is None and not by.startswith('-'):
            remove_list.append(where_list[i])
    for item in remove_list:
        where_list.remove(item)

    where_compare = {}
    for where in where_list:
        compare = {}
        for item in where:
            condition = _create_or_condition(item)
            if compare:
                compare = {
                    'name': 'and',
                    'args': [
                        compare,
                        condition
                    ]
                }
            else:
                compare = condition
        if where_compare:
            where_compare = {
                'name': 'or',
                'args': [
                    where_compare,
                    compare
                ]
            }
        else:
            where_compare = compare

    if where_compare:
        where_compare = asttoexpr(where_compare)

    query = where_compare or None
    return query


def add_page_expr(expr: Expr, page: Page):
    return merge_formulas(expr, asttoexpr({
        'name': 'paginate',
        'args': [
            page
        ]
    }))



import datetime
from typing import List

import sqlalchemy as sa
import tqdm
from typer import echo

from spinta.auth import authorized
from spinta.cli.helpers.errors import ErrorCounter
from spinta.cli.helpers.push import prepare_data_for_push_state
from spinta.cli.helpers.push.utils import extract_state_page_id_key, construct_where_condition_from_page
from spinta.commands.read import PaginationMetaData, get_paginated_values
from spinta.components import Context, Action, Config
from spinta.components import Model
from spinta.components import Page, get_page_size, Property, PageBy
from spinta.utils.response import get_request


def _build_push_state_sync_url(
    server: str,
    model: str,
    page: str,
    page_columns: List[str],
    limit: int
):
    base = f'{server}/{model}/:format/json?'

    required = [
        '_id',
        '_revision',
        '_page',
        'checksum()'
    ]

    base = f'{base}select({",".join(required)}'
    for page_column in page_columns:
        base = f'{base},{page_column}'
    base = f'{base})&sort(_id)'

    if page:
        base = f'{base}&page(\'{page}\')'

    if limit:
        base = f'{base}&limit({limit})'

    return base


def _fetch_all_model_data(
    config: Config,
    model: Model,
    client,
    server: str,
    error_counter: ErrorCounter
):
    limit = config.sync_page_size or 1000
    page = ''
    page_columns = []
    if model.page and model.page.by and model.backend.paginated and model.page.is_enabled:
        for page_by in model.page.by.values():
            page_columns.append(page_by.prop.name)
    while True:
        url = _build_push_state_sync_url(
            server=server,
            model=model.model_type(),
            page=page,
            page_columns=page_columns,
            limit=limit
        )
        status_code, resp = get_request(
            client,
            url,
            error_counter=error_counter,
        )
        if status_code == 200:
            data = resp['_data']
            if not data:
                break

            if '_page' in resp:
                page = resp['_page']['next']

            for row in data:
                yield row

        else:
            break


def _get_state_rows_with_id(
    context: Context,
    table: sa.Table,
    size: int
) -> sa.engine.LegacyCursorResult:
    conn = context.get('push.state.conn')

    model_page = Page()
    model_page.size = size + 1
    prop = Property()
    prop.name = 'id'
    model_page.by = {
        'id': PageBy(prop)
    }
    order_by = sa.sql.expression.nullslast(sa.asc(table.c['id']))

    page_meta = PaginationMetaData(
        page_size=size,
    )

    while not page_meta.is_finished:
        page_meta.is_finished = True
        from_cond = construct_where_condition_from_page(model_page, table, prefix='')

        if from_cond is not None:
            stmt = sa.select([table]).where(
                from_cond
            ).order_by(order_by).limit(model_page.size)
            rows = conn.execute(
                stmt
            )
        else:
            stmt = sa.select([table]).order_by(order_by).limit(model_page.size)
            rows = conn.execute(
                stmt
            )

        yield from get_paginated_values(model_page, page_meta, rows, extract_state_page_id_key)


def _delete_row_from_push_state(
    conn: sa.engine.Connection,
    table: sa.Table,
    id_: str
):
    conn.execute(
        table.delete().
        where(table.c.id == id_)
    )


def _update_row_from_push_state(
    context: Context,
    model: Model,
    conn: sa.engine.Connection,
    table: sa.Table,
    id_: str,
    target_row: dict,
    state_row: dict
):
    target_checksum = target_row.get('checksum()')
    state_checksum = getattr(state_row, 'checksum')

    # Check if target db and pushed data is the same

    page_keys = []
    reserved_keys = [
        '_id',
        '_revision',
        'checksum()'
    ]

    page_mapping = {}
    for key in target_row.keys():
        if key not in reserved_keys:
            page_keys.append(key)

    skip_update = False
    if target_checksum == state_checksum:
        skip_update = True
        # Check if pushed state did not have errors while pushing already existing data
        if getattr(state_row, 'error') or getattr(state_row, 'data'):
            skip_update = False

        # Check if revisions match, they need match in order to do proper updates
        if getattr(state_row, 'revision') != target_row['_revision']:
            skip_update = False

        # Check if page key values match (could be cases when migrating from old to new versions, or changing page keys)
        if skip_update:
            for key in page_keys:
                if target_row[key] != getattr(state_row, f'page.{key}'):
                    skip_update = False
                    break

    # Skip update if nothing has changed
    if skip_update:
        return

    if page_keys:
        page_mapping = {key: target_row[key] for key in page_keys}
        page_mapping = prepare_data_for_push_state(context, model, page_mapping)
        page_mapping = {f'page.{key}': value for key, value in page_mapping.items()}

    conn.execute(
        table.update().
        where(table.c.id == id_).values(
            {
                'revision': target_row.get('_revision'),
                'checksum': target_checksum,
                'pushed': datetime.datetime.now(),
                'error': False,
                'data': None,
                **page_mapping
            }
        )
    )


def _insert_row_to_push_state(
    context: Context,
    model: Model,
    conn: sa.engine.Connection,
    table: sa.Table,
    target_row: dict
):
    page_keys = []
    reserved_keys = [
        '_id',
        '_revision',
        'checksum()'
    ]

    page_mapping = {}
    for key in target_row.keys():
        if key not in reserved_keys:
            page_keys.append(key)

    if page_keys:
        page_mapping = {key: target_row[key] for key in page_keys}
        page_mapping = prepare_data_for_push_state(context, model, page_mapping)
        page_mapping = {f'page.{key}': value for key, value in page_mapping.items()}

    conn.execute(
        table.insert().
        values(
            {
                'id': target_row['_id'],
                'revision': target_row.get('_revision'),
                'checksum': target_row.get('checksum()'),
                'pushed': datetime.datetime.now(),
                'error': False,
                'data': None,
                **page_mapping
            }
        )
    )


def sync_push_state(
    context: Context,
    client,
    server: str,
    models: List[Model],
    error_counter: ErrorCounter,
    no_progress_bar: bool,
    metadata: sa.MetaData
):
    config = context.get('config')
    conn = context.get('push.state.conn')
    counters = {}
    if not no_progress_bar:
        counters = {
            '_total': tqdm.tqdm(desc='SYNCHRONIZING PUSH STATE', ascii=True)
        }
        echo()

    for model in models:
        model_name = model.model_type()
        primary_keys = model.external.pkeys
        size = get_page_size(config, model)
        skip_model = False
        # Check permissions
        for key in primary_keys:
            is_authorized = authorized(context, key, action=Action.SEARCH)
            if not is_authorized:
                echo(f"SKIPPED PUSH STATE '{model.model_type()}' MODEL SYNC, NO PERMISSION.")
                skip_model = True
                break

        if skip_model:
            continue

        if not no_progress_bar:
            counters[model_name] = tqdm.tqdm(desc=model_name, ascii=True)

        model_table = metadata.tables[model.name]
        target_data = _fetch_all_model_data(
            config=config,
            model=model,
            client=client,
            server=server,
            error_counter=error_counter
        )
        state_data = _get_state_rows_with_id(
            context=context,
            table=model_table,
            size=size
        )
        target_row = next(target_data, None)
        state_row = next(state_data, None)
        while target_row is not None or state_row is not None:
            target_id = target_row['_id'] if target_row is not None else None
            state_id = state_row['id'] if state_row is not None else None

            if not no_progress_bar:
                counters['_total'].update(1)
                counters[model_name].update(1)

            if target_row is None:
                _delete_row_from_push_state(conn, model_table, state_id)
                state_row = next(state_data, None)
                continue

            if state_row is None:
                _insert_row_to_push_state(context, model, conn, model_table, target_row)
                target_row = next(target_data, None)
                continue

            if target_id == state_id:
                _update_row_from_push_state(context, model, conn, model_table, state_id, target_row, state_row)
                target_row = next(target_data, None)
                state_row = next(state_data, None)
            elif target_id < state_id:
                _insert_row_to_push_state(context, model, conn, model_table, target_row)
                target_row = next(target_data, None)
            else:
                _delete_row_from_push_state(conn, model_table, state_id)
                state_row = next(state_data, None)

        if model_name in counters:
            counters[model_name].close()

    if '_total' in counters:
        counters['_total'].close()



from typing import List

import geoalchemy2.types
import sqlalchemy as sa

import spinta.backends.postgresql.helpers.migrate.actions as ma
from spinta import commands
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_pg_name
from spinta.backends.postgresql.helpers.migrate.migrate import check_if_renamed, MigratePostgresMeta, \
    adjust_kwargs, extract_literal_name_from_column, handle_unique_constraint_migration, contains_unique_constraint, \
    handle_index_migration
from spinta.components import Context
from spinta.types.datatype import DataType
from spinta.utils.schema import NotAvailable, NA


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, NotAvailable, DataType)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: NotAvailable, new: DataType, **kwargs):
    column = commands.prepare(context, backend, new.prop)
    if column is not None and column != []:
        commands.migrate(context, backend, meta, table, old, column, **kwargs)


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, sa.Column, DataType)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: sa.Column, new: DataType, **kwargs):
    column = commands.prepare(context, backend, new.prop)
    commands.migrate(context, backend, meta, table, old, column, **kwargs)


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, list, DataType)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: List[sa.Column], new: DataType, **kwargs):
    rename = meta.rename

    column: sa.Column = commands.prepare(context, backend, new.prop)
    old_name = rename.get_old_column_name(table.name, column.name)

    columns = old.copy()

    if column.name != old_name:
        for col in old:
            if col.name == column.name:
                commands.migrate(context, backend, meta, table, col, NA, **kwargs)
                columns.remove(col)
                break

    for col in columns:
        commands.migrate(context, backend, meta, table, col, column, **kwargs)


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, sa.Column, sa.Column)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: sa.Column, new: sa.Column, foreign_key: bool = False, **kwargs):
    rename = meta.rename
    inspector = meta.inspector
    handler = meta.handler

    column_name = new.name
    table_name = rename.get_table_name(table.name)
    old_type = extract_literal_name_from_column(old)
    new_type = extract_literal_name_from_column(new)
    using = None

    requires_cast = True
    if isinstance(old.type, geoalchemy2.types.Geometry):
        if isinstance(new.type, geoalchemy2.types.Geometry):
            if old.type.srid != new.type.srid:
                srid_name = old
                srid = new.type.srid
                if old.type.srid == -1:
                    srid_name = sa.func.ST_SetSRID(old, 4326)
                if new.type.srid == -1:
                    srid = 4326
                using = sa.func.ST_Transform(srid_name, srid).compile(compile_kwargs={"literal_binds": True})
                requires_cast = False

    nullable = new.nullable if new.nullable != old.nullable else None
    type_ = new.type if old_type != new_type else None
    new_name = column_name if old.name != new.name else None

    if type_ and requires_cast:
        using = sa.func.cast(old, type_).compile(compile_kwargs={"literal_binds": True})

    if nullable is not None or type_ is not None or new_name is not None or using is not None:
        handler.add_action(ma.AlterColumnMigrationAction(
            table_name=table_name,
            column_name=old.name,
            nullable=nullable,
            type_=type_,
            new_column_name=new_name,
            using=using
        ), foreign_key)
    removed = []
    renamed = check_if_renamed(table.name, table_name, old.name, new.name)
    handle_unique_constraint_migration(
        table,
        table_name,
        old,
        new,
        column_name,
        handler,
        inspector,
        foreign_key,
        removed,
        renamed
    )
    handle_index_migration(
        table,
        table_name,
        old,
        new,
        column_name,
        handler,
        inspector,
        foreign_key,
        removed,
        renamed
    )


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, NotAvailable, sa.Column)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: NotAvailable, new: sa.Column, foreign_key: bool = False, **kwargs):
    rename = meta.rename
    inspector = meta.inspector
    handler = meta.handler

    table_name = get_pg_name(rename.get_table_name(table.name))
    handler.add_action(ma.AddColumnMigrationAction(
        table_name=table_name,
        column=new,
    ), foreign_key)
    if new.unique:
        constraint_name = get_pg_name(f'{table_name}_{new.name}_key')
        if not contains_unique_constraint(inspector, table_name, new.name):
            handler.add_action(ma.CreateUniqueConstraintMigrationAction(
                constraint_name=constraint_name,
                table_name=table_name,
                columns=[new.name]
            ))
    index_required = isinstance(new.type, geoalchemy2.types.Geometry)
    if index_required:
        handler.add_action(ma.CreateIndexMigrationAction(
            table_name=table_name,
            columns=[new.name],
            index_name=get_pg_name(f'idx_{table_name}_{new.name}'),
            using='gist'
        ))


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, list, NotAvailable)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: list, new: NotAvailable, foreign_key: bool = False, **kwargs):
    for item in old:
        commands.migrate(context, backend, meta, table, item, new, **adjust_kwargs(kwargs, 'foreign_key', foreign_key))


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, sa.Column, NotAvailable)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: sa.Column, new: NotAvailable, foreign_key: bool = False, **kwargs):
    rename = meta.rename
    inspector = meta.inspector
    handler = meta.handler

    if not old.name.startswith("_"):
        table_name = rename.get_table_name(table.name)
        columns = inspector.get_columns(table.name)
        remove_name = get_pg_name(f'__{old.name}')

        if any(remove_name == column["name"] for column in columns):
            handler.add_action(ma.DropColumnMigrationAction(
                table_name=table_name,
                column_name=remove_name
            ), foreign_key)
        handler.add_action(ma.AlterColumnMigrationAction(
            table_name=table_name,
            column_name=old.name,
            new_column_name=remove_name
        ), foreign_key)
        indexes = inspector.get_indexes(table_name=table.name)
        for index in indexes:
            if index["column_names"] == [old.name]:
                handler.add_action(ma.DropIndexMigrationAction(
                    table_name=table_name,
                    index_name=index["name"],
                ), foreign_key)
        if old.unique:
            unique_constraints = inspector.get_unique_constraints(table_name=table.name)
            for constraint in unique_constraints:
                if old.name in constraint["column_names"]:
                    handler.add_action(ma.DropConstraintMigrationAction(
                        table_name=table_name,
                        constraint_name=constraint["name"],
                    ), foreign_key)



import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import BIGINT, ARRAY

import spinta.backends.postgresql.helpers.migrate.actions as ma
from spinta import commands
from spinta.backends.components import BackendFeatures
from spinta.backends.constants import TableType
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_pg_name, get_column_name
from spinta.backends.postgresql.helpers.migrate.migrate import get_root_attr, MigratePostgresMeta
from spinta.components import Context
from spinta.types.datatype import File
from spinta.utils.schema import NotAvailable


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, NotAvailable, File)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: NotAvailable, new: File, **kwargs):
    rename = meta.rename
    inspector = meta.inspector
    handler = meta.handler

    name = get_column_name(new.prop)
    nullable = not new.required
    table_name = rename.get_table_name(table.name)
    pkey_type = commands.get_primary_key_type(context, new.backend)
    handler.add_action(ma.AddColumnMigrationAction(
        table_name=table_name,
        column=sa.Column(f'{name}._id', sa.String, nullable=nullable)
    ))
    handler.add_action(ma.AddColumnMigrationAction(
        table_name=table_name,
        column=sa.Column(f'{name}._content_type', sa.String, nullable=nullable)
    ))
    handler.add_action(ma.AddColumnMigrationAction(
        table_name=table_name,
        column=sa.Column(f'{name}._size', BIGINT, nullable=nullable)
    ))
    if BackendFeatures.FILE_BLOCKS in new.backend.features:
        handler.add_action(ma.AddColumnMigrationAction(
            table_name=table_name,
            column=sa.Column(f'{name}._bsize', sa.Integer, nullable=nullable)
        ))
        handler.add_action(ma.AddColumnMigrationAction(
            table_name=table_name,
            column=sa.Column(f'{name}._blocks', ARRAY(pkey_type, ), nullable=nullable)
        ))
    old_table = get_pg_name(f'{table.name}{TableType.FILE.value}/{name}')
    new_table = get_pg_name(f'{table_name}{TableType.FILE.value}/{name}')
    if not inspector.has_table(old_table):
        handler.add_action(ma.CreateTableMigrationAction(
            table_name=new_table,
            columns=[
                sa.Column('_id', pkey_type, primary_key=True),
                sa.Column('_block', sa.LargeBinary)
            ]
        ))


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, list, File)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: list, new: File, **kwargs):
    rename = meta.rename
    inspector = meta.inspector
    handler = meta.handler

    column_name = rename.get_column_name(table.name, new.prop.name)
    old_name = rename.get_old_column_name(table.name, new.prop.name)
    for item in old:
        table_name = rename.get_table_name(table.name)
        nullable = new.required if new.required == item.nullable else None
        item_name = item.name
        new_name = item_name.replace(get_root_attr(item_name), column_name) if not item_name.startswith(
            column_name) else None
        if nullable is not None or new_name is not None:
            handler.add_action(ma.AlterColumnMigrationAction(
                table_name=table_name,
                column_name=item.name,
                nullable=nullable,
                new_column_name=new_name
            ))
    table_name = rename.get_table_name(table.name)
    old_table = get_pg_name(f'{table.name}{TableType.FILE.value}/{old_name}')
    new_table_old_prop = get_pg_name(f'{table_name}{TableType.FILE.value}/{old_name}')
    new_table_new_prop = get_pg_name(f'{table_name}{TableType.FILE.value}/{column_name}')
    if old_name != column_name and inspector.has_table(old_table):
        handler.add_action(
            ma.RenameTableMigrationAction(
                old_table_name=new_table_old_prop,
                new_table_name=new_table_new_prop
            )
        )



from typing import Dict
from typing import Iterable
from typing import List

import sqlalchemy as sa
import tqdm

from spinta.cli.helpers.push.components import PushRow
from spinta.cli.helpers.push.utils import extract_state_page_keys, construct_where_condition_from_page, \
    construct_where_condition_to_page
from spinta.cli.helpers.push.write import prepare_rows_for_deletion
from spinta.commands.read import PaginationMetaData, get_paginated_values
from spinta.components import Context
from spinta.components import Model
from spinta.components import Page, get_page_size


def get_deleted_rows(
    models: List[Model],
    context: Context,
    metadata: sa.MetaData,
    no_progress_bar: bool = False,
):
    counts = (
        _get_deleted_row_counts(
            models,
            context,
            metadata
        )
    )
    total_count = sum(counts.values())
    if total_count > 0:
        rows = _iter_deleted_rows(
            models,
            context,
            metadata,
            counts,
            no_progress_bar
        )
        if not no_progress_bar:
            rows = tqdm.tqdm(rows, 'PUSH DELETED', ascii=True, total=total_count)
        for row in rows:
            yield row


def _get_deleted_row_counts(
    models: List[Model],
    context: Context,
    metadata: sa.MetaData,
) -> dict:
    counts = {}
    conn = context.get('push.state.conn')
    for model in models:
        table = metadata.tables[model.name]
        required_condition = sa.and_(
            table.c.pushed.is_(None),
            table.c.error.is_(False)
        )

        if model.page and model.page.is_enabled and model.page.by:
            where_cond = construct_where_condition_to_page(model.page, table)
        else:
            where_cond = None
        if where_cond is not None:
            where_cond = sa.and_(
                where_cond,
                required_condition
            )
        else:
            where_cond = required_condition

        row_count = conn.execute(
            sa.select(sa.func.count(table.c.id)).
            where(
                where_cond
            )
        )
        counts[model.name] = row_count.scalar()
    return counts


def _iter_deleted_rows(
    models: List[Model],
    context: Context,
    metadata: sa.MetaData,
    counts: Dict[str, int],
    no_progress_bar: bool = False,
) -> Iterable[PushRow]:
    models = reversed(models)
    config = context.get('config')
    conn = context.get('push.state.conn')
    for model in models:
        size = get_page_size(config, model)
        table = metadata.tables[model.name]
        total = counts.get(model.name)

        if model.page and model.page.is_enabled and model.page.by:
            rows = _get_deleted_rows_with_page(context, model.page, table, size)
        else:
            rows = conn.execute(
                sa.select([table.c.id]).
                where(
                    table.c.pushed.is_(None) &
                    table.c.error.is_(False)
                )
            )
        if not no_progress_bar:
            rows = tqdm.tqdm(
                rows,
                model.name,
                ascii=True,
                total=total,
                leave=False
            )

        for row in rows:
            yield prepare_rows_for_deletion(model, row[table.c.id])


def _get_deleted_rows_with_page(
    context: Context,
    model_page: Page,
    table: sa.Table,
    size: int,
) -> sa.engine.LegacyCursorResult:
    conn = context.get('push.state.conn')

    order_by = []

    for page_by in model_page.by.values():
        order_by.append(sa.asc(table.c[f"page.{page_by.prop.name}"]))
    model_page.size = size + 1
    from_page = Page()
    from_page.size = size + 1
    from_page.update_values_from_page(model_page)
    from_page.clear()

    required_where_cond = sa.and_(
        table.c.pushed.is_(None),
        table.c.error.is_(False)
    )

    page_meta = PaginationMetaData(
        page_size=size
    )

    while not page_meta.is_finished:
        page_meta.is_finished = True

        from_cond = construct_where_condition_from_page(from_page, table)
        to_cond = construct_where_condition_to_page(model_page, table)
        where_cond = None

        if from_cond is not None:
            where_cond = from_cond

        if to_cond is not None:
            if where_cond is not None:
                sa.and_(
                    where_cond,
                    to_cond
                )
            else:
                where_cond = to_cond

        if where_cond is not None:
            where_cond = sa.and_(
                required_where_cond,
                where_cond
            )
        else:
            where_cond = required_where_cond

        rows = conn.execute(
            sa.select([table]).
            where(
                where_cond
            ).order_by(*order_by).limit(size)
        )

        yield from get_paginated_values(from_page, page_meta, rows, extract_state_page_keys)








from typing import List

import sqlalchemy as sa
from alembic.migration import MigrationContext
from alembic.operations import Operations
from sqlalchemy.engine.reflection import Inspector

import spinta.backends.postgresql.helpers.migrate.actions as ma
from spinta import commands
from spinta.backends.constants import TableType
from spinta.backends.helpers import get_table_name
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_pg_name, get_column_name
from spinta.backends.postgresql.helpers.migrate.actions import MigrationHandler
from spinta.backends.postgresql.helpers.migrate.migrate import drop_all_indexes_and_constraints, model_name_key, \
    MigratePostgresMeta
from spinta.cli.migrate import MigrateMeta
from spinta.cli.migrate import MigrateRename
from spinta.commands import create_exception
from spinta.components import Context, Model
from spinta.datasets.inspect.helpers import zipitems
from spinta.manifests.components import Manifest
from spinta.types.datatype import Ref, File
from spinta.types.namespace import sort_models_by_ref_and_base
from spinta.utils.schema import NA

EXCLUDED_MODELS = (
    'spatial_ref_sys'
)


@commands.migrate.register(Context, Manifest, PostgreSQL, MigrateMeta)
def migrate(context: Context, manifest: Manifest, backend: PostgreSQL, migrate_meta: MigrateMeta):
    conn = context.get(f'transaction.{backend.name}')
    ctx = MigrationContext.configure(conn, opts={
        "as_sql": migrate_meta.plan,
        "literal_binds": migrate_meta.plan
    })
    op = Operations(ctx)
    inspector = sa.inspect(conn)
    table_names = inspector.get_table_names()
    metadata = sa.MetaData(bind=conn)
    metadata.reflect()

    tables = []
    models = commands.get_models(context, manifest)

    # Filter if only specific dataset can be changed
    if migrate_meta.datasets:
        datasets = migrate_meta.datasets
        filtered_models = {}
        for key, model in models.items():
            if model.external and model.external.dataset and model.external.dataset.name in datasets:
                filtered_models[key] = model
        models = filtered_models

        filtered_names = []
        for table_name in table_names:
            for dataset_name in datasets:
                if table_name.startswith(f'{dataset_name}/'):
                    additional_check = table_name.replace(f'{dataset_name}/', '', 1)
                    if '/' not in additional_check:
                        filtered_names.append(table_name)
        table_names = filtered_names

    for table in table_names:
        name = migrate_meta.rename.get_table_name(table)
        if name not in models.keys():
            name = table
        tables.append(name)
    sorted_models = sort_models_by_ref_and_base(list(models.values()))
    sorted_model_names = list([model.name for model in sorted_models])
    # Do reversed zip, to ensure that sorted models get selected first
    models = zipitems(
        sorted_model_names,
        tables,
        model_name_key
    )
    handler = MigrationHandler()
    meta = MigratePostgresMeta(
        inspector=inspector,
        handler=handler,
        rename=migrate_meta.rename
    )
    for items in models:
        for new_model, old_model in items:
            if old_model and any(value in old_model for value in (TableType.CHANGELOG.value, TableType.FILE.value)):
                continue
            if old_model and old_model in EXCLUDED_MODELS:
                continue
            old = NA
            if old_model:
                old = metadata.tables[migrate_meta.rename.get_old_table_name(old_model)]
            new = commands.get_model(context, manifest, new_model) if new_model else new_model
            commands.migrate(context, backend, meta, old, new)
    _handle_foreign_key_constraints(inspector, sorted_models, handler, migrate_meta.rename)
    _clean_up_file_type(inspector, sorted_models, handler, migrate_meta.rename)
    try:
        if migrate_meta.autocommit:
            # Recreate new connection, that auto commits
            with backend.engine.connect().execution_options(isolation_level="AUTOCOMMIT") as conn:
                ctx = MigrationContext.configure(conn, opts={
                    "as_sql": migrate_meta.plan,
                    "literal_binds": migrate_meta.plan
                })
                op = Operations(ctx)
                handler.run_migrations(op)
        else:
            if migrate_meta.plan:
                with ctx.begin_transaction():
                    handler.run_migrations(op)
            else:
                if ctx._in_connection_transaction():
                    trx = ctx.connection._transaction
                else:
                    trx = ctx.begin_transaction()
                try:
                    handler.run_migrations(op)
                    if migrate_meta.migration_extension is not None:
                        migrate_meta.migration_extension()
                    trx.commit()
                except Exception as e:
                    trx.rollback()
                    raise e
    except sa.exc.OperationalError as error:
        exception = create_exception(manifest, error)
        raise exception


def _handle_foreign_key_constraints(inspector: Inspector, models: List[Model], handler: MigrationHandler,
                                    rename: MigrateRename):
    for model in models:
        source_table = get_pg_name(get_table_name(model))
        old_name = get_pg_name(rename.get_old_table_name(source_table))
        foreign_keys = []
        if old_name in inspector.get_table_names():
            foreign_keys = inspector.get_foreign_keys(old_name)
        if model.base:
            referent_table = get_pg_name(get_table_name(model.base.parent))
            if not model.base.level or model.base.level > 3:
                check = False
                fk_name = get_pg_name(f'fk_{referent_table}_id')
                for key in foreign_keys:
                    if key["constrained_columns"] == ["_id"]:
                        if key["name"] == fk_name and key["referred_table"] == referent_table:
                            check = True
                        else:
                            handler.add_action(ma.DropConstraintMigrationAction(
                                table_name=source_table,
                                constraint_name=key["name"]
                            ), True)
                        break
                if not check:
                    handler.add_action(
                        ma.CreateForeignKeyMigrationAction(
                            source_table=source_table,
                            referent_table=referent_table,
                            constraint_name=fk_name,
                            local_cols=["_id"],
                            remote_cols=["_id"]
                        ), True
                    )
            else:
                for key in foreign_keys:
                    if key["constrained_columns"] == ["_id"]:
                        handler.add_action(ma.DropConstraintMigrationAction(
                            table_name=source_table,
                            constraint_name=key["name"]
                        ), True)
                        break
        else:
            for key in foreign_keys:
                if key["constrained_columns"] == ["_id"]:
                    handler.add_action(ma.DropConstraintMigrationAction(
                        table_name=source_table,
                        constraint_name=key["name"]
                    ), True)
                    break

        required_ref_props = {}
        for prop in model.properties.values():
            if isinstance(prop.dtype, Ref):
                if not prop.level or prop.level > 3:
                    name = get_pg_name(f"fk_{source_table}_{prop.name}._id")
                    required_ref_props[name] = {
                        "name": name,
                        "constrained_columns": [f"{prop.name}._id"],
                        "referred_table": get_pg_name(prop.dtype.model.name),
                        "referred_columns": ["_id"]
                    }

        for key in foreign_keys:
            if key["constrained_columns"] != ["_id"]:
                if key["name"] not in required_ref_props.keys():
                    handler.add_action(ma.DropConstraintMigrationAction(
                        table_name=source_table,
                        constraint_name=key["name"]
                    ), True)
                else:
                    constraint = required_ref_props[key["name"]]
                    if key["constrained_columns"] == constraint["constrained_columns"] and key["referred_table"] == \
                        constraint["referred_table"] and key["referred_columns"] == constraint["referred_columns"]:
                        del required_ref_props[key["name"]]
                    else:
                        handler.add_action(ma.DropConstraintMigrationAction(
                            table_name=source_table,
                            constraint_name=key["name"]
                        ), True)

        for prop in required_ref_props.values():
            handler.add_action(ma.CreateForeignKeyMigrationAction(
                source_table=source_table,
                referent_table=prop["referred_table"],
                constraint_name=prop["name"],
                local_cols=prop["constrained_columns"],
                remote_cols=prop["referred_columns"]
            ), True)


def _clean_up_file_type(inspector: Inspector, models: List[Model], handler: MigrationHandler, rename: MigrateRename):
    allowed_file_tables = []
    existing_tables = []
    for model in models:
        existing_tables.append(rename.get_old_table_name(model.name))
        for prop in model.properties.values():
            if isinstance(prop.dtype, File):
                old_table = rename.get_old_table_name(get_table_name(model))
                old_column = rename.get_old_column_name(old_table, get_column_name(prop))
                allowed_file_tables.append(get_pg_name(f'{old_table}{TableType.FILE.value}/{old_column}'))

    for table in inspector.get_table_names():
        if TableType.FILE.value in table:
            split = table.split(f'{TableType.FILE.value}/')
            if split[0] in existing_tables:
                if table not in allowed_file_tables and not split[1].startswith("__"):
                    new_name = get_pg_name(f'{split[0]}{TableType.FILE.value}/__{split[1]}')
                    if inspector.has_table(new_name):
                        handler.add_action(ma.DropTableMigrationAction(
                            table_name=new_name
                        ))
                    handler.add_action(ma.RenameTableMigrationAction(
                        old_table_name=table,
                        new_table_name=new_name
                    ))
                    drop_all_indexes_and_constraints(inspector, table, new_name, handler)



from typing import Any, List


def ensure_list(value: Any) -> List[Any]:
    if isinstance(value, tuple):
        return list(value)
    if isinstance(value, list):
        return value
    else:
        return [value]



import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSONB

import spinta.backends.postgresql.helpers.migrate.actions as ma
from spinta import commands
from spinta.backends.constants import TableType
from spinta.backends.helpers import get_table_name
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_pg_name
from spinta.backends.postgresql.helpers.migrate.migrate import drop_all_indexes_and_constraints, handle_new_file_type, \
    get_remove_name, get_prop_names, create_changelog_table, property_and_column_name_key, MigratePostgresMeta, \
    MigrateModelMeta, handle_internal_ref_to_scalar_conversion
from spinta.components import Context, Model
from spinta.datasets.inspect.helpers import zipitems
from spinta.types.datatype import File
from spinta.utils.schema import NotAvailable, NA


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, Model)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, old: sa.Table, new: Model, **kwargs):
    rename = meta.rename
    handler = meta.handler
    inspector = meta.inspector

    columns = list(old.columns)
    table_name = rename.get_table_name(old.name)

    if table_name != old.name:
        handler.add_action(ma.RenameTableMigrationAction(old.name, table_name))
        changelog_name = get_pg_name(f'{old.name}{TableType.CHANGELOG.value}')
        file_name = get_pg_name(f'{old.name}{TableType.FILE.value}')
        if inspector.has_table(changelog_name):
            new_changelog_name = get_pg_name(f'{table_name}{TableType.CHANGELOG.value}')
            handler.add_action(ma.RenameTableMigrationAction(
                old_table_name=changelog_name,
                new_table_name=new_changelog_name
            ))
            handler.add_action(ma.RenameSequenceMigrationAction(
                old_name=get_pg_name(f'{changelog_name}__id_seq'),
                new_name=get_pg_name(f'{new_changelog_name}__id_seq')
            ))
        for table in inspector.get_table_names():
            if table.startswith(file_name):
                split = table.split(TableType.FILE.value)
                handler.add_action(ma.RenameTableMigrationAction(
                    old_table_name=file_name,
                    new_table_name=get_pg_name(f'{table_name}{TableType.FILE.value}{split[1]}')
                ))

    properties = list(new.properties.values())

    model_meta = MigrateModelMeta()
    for column in columns:
        if isinstance(column.type, JSONB):
            model_meta.add_json_column(
                backend,
                old,
                column
            )

    props = zipitems(
        columns,
        properties,
        lambda x: property_and_column_name_key(x, rename, old, new)
    )
    for items in props:
        old_columns = []
        new_properties = []
        for old_prop, new_prop in items:
            if new_prop and new_prop.name.startswith('_'):
                continue

            if old_prop not in old_columns:
                old_columns.append(old_prop)
            if new_prop and new_prop not in new_properties:
                new_properties.append(new_prop)

        if len(old_columns) == 1:
            old_columns = old_columns[0]
        elif not old_columns:
            old_columns = NA
            if not new_properties:
                continue

        if new_properties:
            for prop in new_properties:
                if not prop:
                    commands.migrate(context, backend, meta, old, old_columns, prop, model_meta=model_meta, foreign_key=False, **kwargs)
                else:
                    handled = handle_internal_ref_to_scalar_conversion(
                        context,
                        backend,
                        meta,
                        old,
                        old_columns,
                        prop
                    )
                    if not handled:
                        commands.migrate(context, backend, meta, old, old_columns, prop, model_meta=model_meta, **kwargs)
        else:
            # Remove from JSONB clean up
            if isinstance(old_columns, sa.Column) and isinstance(old_columns.type, JSONB):
                columns.remove(old_columns)
            commands.migrate(context, backend, meta, old, old_columns, NA, model_meta=model_meta, foreign_key=False, **kwargs)

    required_unique_constraints = []
    if new.unique:
        for val in new.unique:
            prop_list = []

            for prop in val:
                for name in get_prop_names(prop):
                    prop_list.append(get_pg_name(name))

            constraints = inspector.get_unique_constraints(old.name)
            constraint_name = f'{old.name}_{"_".join(prop_list)}_key'
            if old.name != table_name:
                constraint_name = f'{table_name}_{"_".join(prop_list)}_key'
            constraint_name = get_pg_name(constraint_name)
            required_unique_constraints.append(prop_list)

            # Skip unique properties, since they are already handled in other parts
            if not (len(val) == 1 and val[0].dtype.unique):
                if not any(constraint['name'] == constraint_name for constraint in constraints):
                    handler.add_action(ma.CreateUniqueConstraintMigrationAction(
                        table_name=table_name,
                        constraint_name=constraint_name,
                        columns=prop_list
                    ))

    for prop in new.properties.values():
        if prop.dtype.unique:
            result = commands.prepare(context, backend, prop.dtype)
            if not isinstance(result, list):
                result = [result]
            prop_list = []
            for item in result:
                if isinstance(item, sa.Column):
                    prop_list.append(item.name)
            if prop_list:
                required_unique_constraints.append(prop_list)

    # Clean up old multi constraints for non required models (happens when model.ref is changed)
    # Solo unique constraints are handled with property migrate
    if not old.name.startswith("_"):
        unique_constraints = inspector.get_unique_constraints(old.name)
        for constraint in unique_constraints:
            if constraint["column_names"] not in required_unique_constraints:
                if not handler.has_constraint_been_dropped(constraint["name"]):
                    handler.add_action(ma.DropConstraintMigrationAction(
                        table_name=table_name,
                        constraint_name=constraint["name"]
                    ))

    # Handle JSON migrations, that need to be run at the end
    for json_meta in model_meta.json_columns.values():
        if json_meta.new_keys and json_meta.cast_to is None:
            removed_keys = [key for key, new_key in json_meta.new_keys.items() if new_key == f'__{key}']

            if removed_keys == json_meta.keys or json_meta.full_remove:
                commands.migrate(context, backend, meta, old, json_meta.column, NA)
            else:
                for old_key, new_key in json_meta.new_keys.items():
                    if new_key in json_meta.keys:
                        handler.add_action(ma.RemoveJSONAttributeMigrationAction(
                            table_name,
                            json_meta.column,
                            new_key
                        ))
                    handler.add_action(ma.RenameJSONAttributeMigrationAction(
                        table_name,
                        json_meta.column,
                        old_key,
                        new_key
                    ))
        elif isinstance(json_meta.cast_to, tuple) and len(json_meta.cast_to) == 2:
            new_column = json_meta.cast_to[0]
            key = json_meta.cast_to[1]
            commands.migrate(context, backend, meta, old, json_meta.column, NA)
            commands.migrate(context, backend, meta, old, NA, new_column, foreign_key=False, model_meta=model_meta, **kwargs)
            renamed = json_meta.column._copy()
            renamed.name = f'__{json_meta.column.name}'
            renamed.key = f'__{json_meta.column.name}'
            handler.add_action(
                ma.TransferJSONDataMigrationAction(
                    table_name,
                    renamed,
                    [(key, new_column)]
                )
            )
        # Rename column
        if json_meta.new_name:
            handler.add_action(ma.AlterColumnMigrationAction(
                table_name,
                json_meta.column.name,
                new_column_name=json_meta.new_name
            ))


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, NotAvailable, Model)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, old: NotAvailable, new: Model, **kwargs):
    rename = meta.rename
    handler = meta.handler
    inspector = meta.inspector

    table_name = get_pg_name(get_table_name(new))
    pkey_type = commands.get_primary_key_type(context, backend)

    columns = []
    for prop in new.properties.values():
        if not prop.name.startswith("_"):
            if isinstance(prop.dtype, File):
                columns += handle_new_file_type(context, backend, inspector, prop, pkey_type, handler)
            else:
                cols = commands.prepare(context, backend, prop)
                if isinstance(cols, list):
                    for column in cols:
                        if isinstance(column, (sa.Column, sa.Index)):
                            columns.append(column)
                else:
                    if isinstance(cols, sa.Column):
                        columns.append(cols)

    handler.add_action(ma.CreateTableMigrationAction(
        table_name=table_name,
        columns=[
            sa.Column('_txn', pkey_type, index=True),
            sa.Column('_created', sa.DateTime),
            sa.Column('_updated', sa.DateTime),
            sa.Column("_id", pkey_type, nullable=False, primary_key=True),
            sa.Column("_revision", sa.Text),
            *columns
        ]

    ))

    if new.unique:
        for val in new.unique:
            prop_list = []
            for prop in val:
                for name in get_prop_names(prop):
                    prop_list.append(get_pg_name(name))
            constraint_name = get_pg_name(f'{table_name}_{"_".join(prop_list)}_key')
            handler.add_action(ma.CreateUniqueConstraintMigrationAction(
                table_name=table_name,
                constraint_name=constraint_name,
                columns=prop_list
            ))

    create_changelog_table(context, new, handler, rename)


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, NotAvailable)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, old: sa.Table, new: NotAvailable):
    handler = meta.handler
    inspector = meta.inspector

    old_name = old.name
    remove_name = get_remove_name(old_name)

    if not old_name.split("/")[-1].startswith("__"):
        if inspector.has_table(remove_name):
            handler.add_action(ma.DropTableMigrationAction(
                table_name=remove_name
            ))
            remove_changelog_name = get_pg_name(f'{remove_name}{TableType.CHANGELOG.value}')
            if inspector.has_table(remove_changelog_name):
                handler.add_action(ma.DropTableMigrationAction(
                    table_name=remove_changelog_name
                ))
            remove_file_name = get_pg_name(f'{remove_name}{TableType.FILE.value}')
            for table in inspector.get_table_names():
                if table.startswith(remove_file_name):
                    handler.add_action(ma.DropTableMigrationAction(
                        table_name=table
                    ))

        handler.add_action(ma.RenameTableMigrationAction(
            old_table_name=old_name,
            new_table_name=remove_name
        ))
        drop_all_indexes_and_constraints(inspector, old_name, remove_name, handler)

        old_changelog_name = get_pg_name(f'{old_name}{TableType.CHANGELOG.value}')
        new_changelog_name = get_pg_name(f'{remove_name}{TableType.CHANGELOG.value}')
        if inspector.has_table(old_changelog_name):
            handler.add_action(ma.RenameTableMigrationAction(
                old_table_name=old_changelog_name,
                new_table_name=new_changelog_name
            ))
            handler.add_action(ma.RenameSequenceMigrationAction(
                old_name=f'{old_changelog_name}__id_seq',
                new_name=f'{new_changelog_name}__id_seq'
            ))
            drop_all_indexes_and_constraints(inspector, old_changelog_name, new_changelog_name, handler)
        for table in inspector.get_table_names():
            if table.startswith(f'{old_name}{TableType.FILE.value}'):
                split = table.split(TableType.FILE.value)
                new_name = get_pg_name(f'{remove_name}{TableType.FILE.value}{split[1]}')
                handler.add_action(ma.RenameTableMigrationAction(
                    old_table_name=table,
                    new_table_name=new_name
                ))
                drop_all_indexes_and_constraints(inspector, table, new_name, handler)






from typing import List

import sqlalchemy as sa

import spinta.backends.postgresql.helpers.migrate.actions as ma
from spinta import commands
from spinta.backends.helpers import get_table_name
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_column_name
from spinta.backends.postgresql.helpers.migrate.migrate import name_key, MigratePostgresMeta, is_internal_ref, \
    adjust_kwargs, is_name_complex, extract_literal_name_from_column, generate_type_missmatch_exception_details
from spinta.components import Context
from spinta.datasets.inspect.helpers import zipitems
from spinta.exceptions import MigrateScalarToRefTooManyKeys, MigrateScalarToRefTypeMissmatch
from spinta.types.datatype import Ref
from spinta.utils.schema import NotAvailable, NA


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, NotAvailable, Ref)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: NotAvailable, new: Ref, **kwargs):
    columns = commands.prepare(context, backend, new.prop)
    if not isinstance(columns, list):
        columns = [columns]
    for column in columns:
        if isinstance(column, sa.Column):
            commands.migrate(context, backend, meta, table, old, column, **adjust_kwargs(kwargs, 'foreign_key', True))


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, sa.Column, Ref)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: sa.Column, new: Ref, **kwargs):
    commands.migrate(context, backend, meta, table, [old], new, **kwargs)


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, list, Ref)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: List[sa.Column], new: Ref, **kwargs):
    rename = meta.rename
    inspector = meta.inspector
    handler = meta.handler

    adjusted_kwargs = adjust_kwargs(kwargs, 'foreign_key', True)

    new_columns = commands.prepare(context, backend, new.prop)
    if not isinstance(new_columns, list):
        new_columns = [new_columns]
    table_name = rename.get_table_name(table.name)
    old_ref_table = rename.get_old_table_name(get_table_name(new.model))
    old_prop_name = rename.get_old_column_name(table.name, get_column_name(new.prop))

    old_names = {}
    new_names = {}
    for item in old:
        base_name = item.name.split(".")
        name = rename.get_column_name(table.name, item.name)
        if len(base_name) > 1:
            name = f'{rename.get_column_name(table.name, base_name[0])}.{rename.get_column_name(old_ref_table, base_name[1])}'
        old_names[name] = item

    for item in new_columns:
        new_names[item.name] = item

    if is_internal_ref(new):
        column = new_columns[0]
        if len(old) == 1 and old[0].name == f'{old_prop_name}._id':
            old_column = old[0]
            commands.migrate(context, backend, meta, table, old_column, column, **adjusted_kwargs)
        else:
            column_list = dict()
            drop_list = []
            for old_column in old:
                new_name = rename.get_column_name(table.name, old_column.name)
                key = new_name.split('.')[-1]

                # Check for scalar -> ref 4, can only accept 1 primary key
                if key == new_name:
                    model_external = new.model.external
                    key = '_id'
                    if model_external and not model_external.unknown_primary_key and model_external.pkeys:
                        if len(model_external.pkeys) > 1:
                            raise MigrateScalarToRefTooManyKeys(new, primary_keys=[pkey.name for pkey in model_external.pkeys])
                        target = model_external.pkeys[0]
                        target_column = commands.prepare(context, backend, target)
                        key = target.name
                        old_type = extract_literal_name_from_column(old_column)
                        new_type = extract_literal_name_from_column(target_column)
                        if old_type != new_type:
                            raise MigrateScalarToRefTypeMissmatch(new,
                                  details=generate_type_missmatch_exception_details(
                                      [
                                          (
                                              (old_column.name, old_type),
                                              (key, new_type)
                                          )
                                      ]
                                  ))

                column_list[key] = sa.Column(new_name, old_column.type)
                if old_column.name != f'{old_prop_name}._id':
                    drop_list.append(old_column)

            inspector_columns = inspector.get_columns(table.name)
            old_col = NA
            if any(f'{old_prop_name}._id' == col["name"] for col in inspector_columns):
                old_col = table.c.get(f'{old_prop_name}._id')
            commands.migrate(context, backend, meta, table, old_col, column, **adjusted_kwargs)
            if old_names.keys() != new_names.keys():
                handler.add_action(ma.UpgradeTransferDataMigrationAction(
                    table_name=table_name,
                    foreign_table_name=get_table_name(new.refprops[0]),
                    target=column,
                    columns=column_list
                ), True)
            for col in drop_list:
                commands.migrate(context, backend, meta, table, col, NA, **adjusted_kwargs)
    else:
        old_name_with_id = f'{old_prop_name}._id'
        # Might going to need to add better check to see if its transform or rename
        # This will prob not work with nested structure
        if len(old) == 1 and (
            old[0].name == old_name_with_id or
            not is_name_complex(old[0].name)
        ):
            column = old[0]
            requires_drop = True
            if len(new_columns) > 1 and column.name != old_name_with_id:
                raise MigrateScalarToRefTooManyKeys(new, primary_keys=[col.name.split('.')[-1] for col in new_columns])

            target = '_id'

            # Scalar to ref checks and adjustments
            if not column.name.endswith('._id') and len(new_columns) == 1 and isinstance(new_columns[0], sa.Column):
                target_column = new_columns[0]
                target = target_column.name.split('.')[-1]
                old_type = extract_literal_name_from_column(column)
                new_type = extract_literal_name_from_column(target_column)
                if old_type != new_type:
                    raise MigrateScalarToRefTypeMissmatch(new, details=generate_type_missmatch_exception_details(
                        [
                            (
                                (column.name, old_type),
                                (target_column.name, new_type)
                            )
                        ]
                    ))

            for new_column in new_columns:
                if isinstance(new_column, sa.Column):
                    if new_column.name == old_name_with_id:
                        requires_drop = False
                        commands.migrate(context, backend, meta, table, old[0], new_column, **adjusted_kwargs)
                    else:
                        commands.migrate(context, backend, meta, table, NA, new_column, **adjusted_kwargs)
            column_list = dict()

            for item in new_columns:
                name = item.name
                column_list[name] = sa.Column(name.split('.')[-1], item.type)
            if old_names.keys() != new_names.keys():
                handler.add_action(
                    ma.DowngradeTransferDataMigrationAction(
                        table_name=table_name,
                        foreign_table_name=get_table_name(new.refprops[0]),
                        source=column,
                        columns=column_list,
                        target=target
                    ), True
                )
            if requires_drop:
                commands.migrate(context, backend, meta, table, old[0], NA, **adjusted_kwargs)
        else:
            props = zipitems(
                old_names.keys(),
                new_names.keys(),
                name_key
            )
            drop_list = []
            for prop in props:
                for old_prop, new_prop in prop:
                    if old_prop is not NA:
                        old_prop = old_names[old_prop]
                    if new_prop is not NA:
                        new_prop = new_names[new_prop]
                    if old_prop is not None and new_prop is None:
                        drop_list.append(old_prop)
                    else:
                        commands.migrate(context, backend, meta, table, old_prop, new_prop, **adjusted_kwargs)
            for drop in drop_list:
                commands.migrate(context, backend, meta, table, drop, NA, **adjusted_kwargs)



import dataclasses
from typing import Any, List, Union, Dict, Tuple

import sqlalchemy as sa
import geoalchemy2.types
from sqlalchemy.dialects.postgresql import JSONB, BIGINT, ARRAY, JSON
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.dialects import postgresql

import spinta.backends.postgresql.helpers.migrate.actions as ma
from spinta import commands
from spinta.backends.components import BackendFeatures
from spinta.backends.constants import TableType
from spinta.backends.helpers import get_table_name
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_pg_name, get_column_name
from spinta.backends.postgresql.helpers.migrate.actions import MigrationHandler
from spinta.cli.migrate import MigrateRename
from spinta.components import Context, Model, Property
from spinta.datasets.enums import Level
from spinta.exceptions import MigrateScalarToRefTooManyKeys
from spinta.types.datatype import Ref, File, Array, Object
from spinta.types.text.components import Text
from spinta.utils.nestedstruct import get_root_attr
from spinta.utils.schema import NA


def check_if_renamed(old_table: str, new_table: str, old_property: str, new_property: str):
    return old_table != new_table or old_property != new_property


def drop_all_indexes_and_constraints(inspector: Inspector, table: str, new_table: str, handler: MigrationHandler):
    constraints = inspector.get_unique_constraints(table)
    removed = []
    foreign_keys = inspector.get_foreign_keys(table)
    for key in foreign_keys:
        handler.add_action(ma.DropConstraintMigrationAction(
            table_name=new_table,
            constraint_name=key["name"]
        ), True)

    for constraint in constraints:
        removed.append(constraint["name"])
        handler.add_action(
            ma.DropConstraintMigrationAction(
                table_name=new_table,
                constraint_name=constraint["name"]
            )
        )
    indexes = inspector.get_indexes(table)
    for index in indexes:
        if index["name"] not in removed:
            handler.add_action(
                ma.DropIndexMigrationAction(
                    table_name=new_table,
                    index_name=index["name"]
                )
            )


def create_changelog_table(context: Context, new: Model, handler: MigrationHandler, rename: MigrateRename):
    table_name = get_pg_name(get_table_name(new, TableType.CHANGELOG))
    pkey_type = commands.get_primary_key_type(context, new.backend)
    handler.add_action(ma.CreateTableMigrationAction(
        table_name=table_name,
        columns=[
            sa.Column('_id', BIGINT, primary_key=True, autoincrement=True),
            sa.Column('_revision', sa.String),
            sa.Column('_txn', pkey_type, index=True),
            sa.Column('_rid', pkey_type),
            sa.Column('datetime', sa.DateTime),
            sa.Column('action', sa.String(8)),
            sa.Column('data', JSONB)
        ]
    ))


def handle_new_file_type(context: Context, backend: PostgreSQL, inspector: Inspector, prop: Property, pkey_type: Any,
                         handler: MigrationHandler) -> list:
    name = get_column_name(prop)
    nullable = not prop.dtype.required
    columns = []
    columns += [
        sa.Column(f'{name}._id', sa.String, nullable=nullable),
        sa.Column(f'{name}._content_type', sa.String, nullable=nullable),
        sa.Column(f'{name}._size', BIGINT, nullable=nullable)
    ]
    if BackendFeatures.FILE_BLOCKS in prop.dtype.backend.features:
        columns += [
            sa.Column(f'{name}._bsize', sa.Integer, nullable=nullable),
            sa.Column(f'{name}._blocks', ARRAY(pkey_type, ), nullable=nullable),
        ]
    new_table = get_pg_name(get_table_name(prop, TableType.FILE))
    if not inspector.has_table(new_table):
        handler.add_action(ma.CreateTableMigrationAction(
            table_name=new_table,
            columns=[
                sa.Column('_id', pkey_type, primary_key=True),
                sa.Column('_block', sa.LargeBinary)
            ]
        ))
    return columns


def handle_new_array_type(context: Context, backend: PostgreSQL, inspector: Inspector, prop: Property, pkey_type: Any,
                          handler: MigrationHandler):
    columns = []
    if isinstance(prop.dtype, Array) and prop.dtype.items:
        if prop.list is None:
            columns.append(sa.Column(prop.place, JSONB))

        if isinstance(prop.dtype.items.dtype, File):
            new_columns = handle_new_file_type(context, backend, inspector, prop.dtype.items, pkey_type, handler)
        elif isinstance(prop.dtype.items.dtype, Array):
            new_columns = handle_new_array_type(context, backend, inspector, prop.dtype.items, pkey_type, handler)
        elif isinstance(prop.dtype.items.dtype, Object):
            new_columns = handle_new_object_type(context, backend, inspector, prop.dtype.items, pkey_type, handler)
        else:
            new_columns = commands.prepare(context, backend, prop.dtype.items)
        if not isinstance(new_columns, list):
            new_columns = [new_columns]
        for column in new_columns:
            if not isinstance(column, sa.Column):
                new_columns.remove(column)

        new_table = get_pg_name(get_table_name(prop, TableType.LIST))
        if not inspector.has_table(new_table):
            main_table_name = get_pg_name(get_table_name(prop.model))
            handler.add_action(ma.CreateTableMigrationAction(
                table_name=new_table,
                columns=[
                    sa.Column('_txn', pkey_type, index=True),
                    sa.Column('_rid', pkey_type, sa.ForeignKey(
                        f'{main_table_name}._id', ondelete='CASCADE',
                    ), index=True),
                    *new_columns
                ]
            ))
    return columns


def handle_new_object_type(context: Context, backend: PostgreSQL, inspector: Inspector, prop: Property, pkey_type: Any,
                           handler: MigrationHandler):
    columns = []
    if isinstance(prop.dtype, Object) and prop.dtype.properties:
        for new_prop in prop.dtype.properties.values():
            if prop.name.startswith('_') and prop.name not in ('_revision',):
                continue
            if isinstance(new_prop.dtype, File):
                columns = handle_new_file_type(context, backend, inspector, new_prop, pkey_type, handler)
            elif isinstance(new_prop.dtype, Array):
                columns = handle_new_array_type(context, backend, inspector, new_prop, pkey_type, handler)
            elif isinstance(new_prop.dtype, Object):
                columns = handle_new_object_type(context, backend, inspector, new_prop, pkey_type, handler)
            else:
                columns = commands.prepare(context, backend, new_prop)

            if not isinstance(columns, list):
                columns = [columns]
            for column in columns:
                if not isinstance(column, sa.Column):
                    columns.remove(column)
    return columns


def rename_index_name(index: str, old_table: str, new_table: str, old_property: str, new_property: str):
    new = f'ix_{new_table}_{new_property}'
    return new


def get_remove_name(name: str) -> str:
    new_name = name.split("/")
    if not new_name[-1].startswith("__"):
        new_name[-1] = f'__{new_name[-1]}'
    new_name = '/'.join(new_name)
    new_name = get_pg_name(new_name)
    return new_name


def get_prop_names(prop: Property):
    name = prop.name
    if isinstance(prop.dtype, Ref):
        if not prop.level or prop.level > Level.open:
            name = f'{name}._id'
        else:
            for refprop in prop.dtype.refprops:
                yield f'{name}.{refprop.name}'
    yield name


def json_has_key(backend: PostgreSQL, column: sa.Column, table: sa.Table, key: str):
    with backend.engine.begin() as connection:
        query = sa.select(table.select().where(column.has_key(key)).exists())
        return connection.execute(query).scalar()


def jsonb_keys(backend: PostgreSQL, column: sa.Column, table: sa.Table):
    with backend.engine.begin() as connection:
        keys = sa.func.jsonb_object_keys(
            column
        )
        query = sa.select(
            [keys]
        ).select_from(table).group_by(keys)
        return [result[0] for result in connection.execute(query)]


def name_key(name: str):
    return name


def model_name_key(model: str) -> str:
    return get_pg_name(model)


def is_name_complex(name: str):
    return '.' in name or '@' in name


def is_prop_complex(prop: Property):
    return isinstance(prop.dtype, (Text, File, Ref))


def is_name_or_property_complex(name: str, prop: Property):
    is_complex = is_name_complex(name)
    if not is_complex:
        return is_prop_complex(prop)
    return is_complex


def is_name_or_column_complex(name: str, col: sa.Column):
    is_complex = is_name_complex(name)
    if not is_complex:
        return is_column_complex(col)
    return is_complex


def is_column_complex(col: sa.Column):
    return isinstance(col.type, (JSONB, JSON))


def property_and_column_name_key(item: Union[sa.Column, Property], rename, table: sa.Table, model: Model) -> str:
    # Mapping concept is to prioritize complex types over simple
    # new types take priority over old
    # Column is always old, Property is always new

    if isinstance(item, sa.Column):
        name = item.name
        is_complex = is_name_or_column_complex(name, item)
        new_name = rename.get_column_name(table.name, name, True)
        full_name = rename.get_column_name(table.name, name)

        root_changed = has_been_renamed(name, new_name)
        full_changed = has_been_renamed(name, full_name)

        # Check for edge case when you have old columns: column_one, column_two
        # new manifest only hase column_one, but
        # rename provides "column_two": "column_one"
        # meaning, you need to remove old "column_one" and rename old "column_two" to "column_one"
        if not root_changed:
            old_name = rename.get_old_column_name(table.name, name)
            if has_been_renamed(name, old_name):
                return get_root_attr(old_name)

        if full_changed:
            new_prop = model.flatprops[full_name]
            if is_name_or_property_complex(full_name, new_prop) or is_complex:
                return get_root_attr(full_name)

        if root_changed:
            new_prop = model.flatprops[new_name]
            if is_name_or_property_complex(new_name, new_prop):
                return get_root_attr(new_name)
            elif not full_changed:
                return name

        return get_root_attr(name)
    elif isinstance(item, Property):
        name = get_column_name(item)
        old_name = rename.get_old_column_name(table.name, name, True)
        old_full_name = rename.get_old_column_name(table.name, name)

        if is_name_or_property_complex(name, item):
            return get_root_attr(name)

        if has_been_renamed(name, old_full_name):
            if old_full_name in table.columns:
                col = table.columns[old_full_name]
                if is_name_or_column_complex(old_full_name, col):
                    return get_root_attr(name)

        return get_root_attr(old_name)


@dataclasses.dataclass
class MigratePostgresMeta:
    inspector: Inspector
    rename: MigrateRename
    handler: MigrationHandler


@dataclasses.dataclass
class JSONColumnMigrateMeta:
    column: sa.Column
    keys: List[str] = dataclasses.field(default_factory=list)
    new_keys: Dict[str, str] = dataclasses.field(default_factory=dict)
    full_remove: bool = dataclasses.field(default=True)
    cast_to: Tuple[sa.Column, str] = dataclasses.field(default=None)
    new_name: str = dataclasses.field(default=None)

    def initialize(self, backend, table):
        self.keys = jsonb_keys(backend, self.column, table)

    def add_new_key(self, old_key: str, new_key: str):
        if old_key in self.keys and old_key not in self.new_keys:
            self.new_keys[old_key] = new_key


@dataclasses.dataclass
class MigrateModelMeta:
    json_columns: Dict[str, JSONColumnMigrateMeta] = dataclasses.field(default_factory=dict)

    def add_json_column(self, backend: PostgreSQL, table: sa.Table, column: sa.Column):
        meta = JSONColumnMigrateMeta(
            column=column,
        )
        meta.initialize(backend, table)
        self.json_columns[column.name] = meta


def has_been_renamed(old: str, new: str):
    return old != new


def is_internal_ref(dtype: Ref):
    prop = dtype.prop
    return prop.level is None or prop.level > Level.open


def handle_internal_ref_to_scalar_conversion(
    context: Context,
    backend: PostgreSQL,
    meta: MigratePostgresMeta,
    table: sa.Table,
    old_columns: List[sa.Column],
    new_property: Property,
    **kwargs
) -> bool:
    if isinstance(old_columns, sa.Column):
        old_columns = [old_columns]

    if not old_columns or not new_property:
        return False

    # Check if columns are from ref 4 (can only have 1 column)
    if not (len(old_columns) == 1 and isinstance(old_columns[0], sa.Column)):
        return False

    inspector = meta.inspector
    handler = meta.handler
    rename = meta.rename

    ref_col = old_columns[0]
    manifest = new_property.model.manifest

    # Skip ref 4 -> ref 3
    if isinstance(new_property.dtype, Ref):
        return False

    if ref_col.name.endswith('._id'):
        constraints = inspector.get_foreign_keys(table.name)
        ref_model = None
        for constraint in constraints:
            if constraint['constrained_columns'] == [ref_col.name]:
                table_name = constraint['referred_table']
                if commands.has_model(context, manifest, table_name):
                    ref_model = commands.get_model(context, manifest, table_name)
                else:
                    # In case table name has been truncated, need to loop through all models and convert their names to pg
                    all_models = commands.get_models(context, manifest)
                    for model in all_models.values():
                        if get_pg_name(model.name) == table_name:
                            ref_model = model
                            break
                break

        if isinstance(ref_model, Model):
            if ref_model.external and not ref_model.external.unknown_primary_key:
                pkeys = ref_model.external.pkeys
                mapped_data: dict = dict()
                if len(pkeys) > 1:
                    raise MigrateScalarToRefTooManyKeys(new_property.dtype, primary_keys=[key.name for key in pkeys])

                target_key = pkeys[0]
                prop_col = commands.prepare(context, backend, target_key)
                mapped_data[get_pg_name(new_property.name)] = prop_col
                updated_kwargs = adjust_kwargs(kwargs, 'foreign_key', True)

                commands.migrate(context, backend, meta, table, NA, new_property, **updated_kwargs)
                table_name = rename.get_table_name(table.name)
                foreign_table_name = get_table_name(ref_model)
                handler.add_action(
                    ma.DowngradeTransferDataMigrationAction(
                        table_name,
                        foreign_table_name,
                        ref_col,
                        mapped_data
                    ),
                    foreign_key=True
                )
                commands.migrate(context, backend, meta, table, ref_col, NA, **updated_kwargs)
                return True

    return False


def extract_target_column(rename: MigrateRename, columns: list, table: sa.Table, prop: Property):
    full_name = rename.get_old_column_name(table.name, prop.name)
    if isinstance(columns, list):
        for col in columns:
            if isinstance(col, sa.Column) and col.name == full_name:
                return [col]
    return columns


def adjust_kwargs(kwargs: dict, key: str, value: Any) -> dict:
    copied = kwargs.copy()
    copied[key] = value
    return copied


def extract_literal_name_from_column(
    column: sa.Column,
) -> str:
    type_ = column.type.compile(dialect=postgresql.dialect())

    # Convert sa.Float, to postgresql DOUBLE PRECISION type
    if isinstance(column.type, sa.Float):
        type_ = 'DOUBLE PRECISION'

    return type_


# Match [
#   (
#       (old_column_name, old_type),
#       (new_column_name, new_type)
#   )
# ]
def generate_type_missmatch_exception_details(
    columns: list
):
    result = ''
    for pair in columns:
        old_data = pair[0]
        new_data = pair[1]
        result += f'\t\'{old_data[0]}\' [{old_data[1]}] -> \'{new_data[0]}\' [{new_data[1]}]\t'
        if old_data[1] == new_data[1]:
            result += f'\'{old_data[1]}\' == \'{new_data[1]}\'\n'
        else:
            result += f'\'{old_data[1]}\' != \'{new_data[1]}\'\t<= Incorrect\n'
    return result


def contains_unique_constraint(inspector: Inspector, table: Union[sa.Table, str], column_name: str):
    table_name = table.name if isinstance(table, sa.Table) else table
    return any(constraint["column_names"] == [column_name] for constraint in
               inspector.get_unique_constraints(table_name=table_name)) or any(
        index["column_names"] == [column_name] and index["unique"] for index in
        inspector.get_indexes(table_name=table_name))


def handle_unique_constraint_migration(
    table: sa.Table,
    table_name: str,
    old: sa.Column,
    new: sa.Column,
    column_name: str,
    handler: MigrationHandler,
    inspector: Inspector,
    foreign_key: bool,
    removed: list,
    renamed: bool
):
    unique_name = get_pg_name(f'{table_name}_{column_name}_key')

    unique_constraints = inspector.get_unique_constraints(table_name=table.name)
    constraint_column = old.name if renamed else column_name

    if new.unique and not renamed and not contains_unique_constraint(inspector, table, constraint_column):
        handler.add_action(ma.CreateUniqueConstraintMigrationAction(
            constraint_name=unique_name,
            table_name=table_name,
            columns=[constraint_column]
        ), foreign_key)
    else:
        for constraint in unique_constraints:
            dropped = False
            if constraint["column_names"] == [constraint_column]:
                removed.append(constraint["name"])
                if renamed or old.unique:
                    dropped = True
                    handler.add_action(ma.DropConstraintMigrationAction(
                        constraint_name=constraint["name"],
                        table_name=table_name
                    ), foreign_key)
                unique_name = get_pg_name(
                    rename_index_name(constraint["name"], table.name, table_name, old.name, new.name))
                if new.unique and (dropped or not contains_unique_constraint(inspector, table, constraint_column)):
                    handler.add_action(ma.CreateUniqueConstraintMigrationAction(
                        constraint_name=unique_name,
                        table_name=table_name,
                        columns=[constraint_column]
                    ), foreign_key)
        if not new.unique:
            for index in inspector.get_indexes(table_name=table.name):
                if index["column_names"] == [column_name] and index["unique"] and index["name"] not in removed:
                    index_name = index["name"]
                    handler.add_action(ma.DropIndexMigrationAction(
                        index_name=index_name,
                        table_name=table_name,
                    ), foreign_key)
                    handler.add_action(ma.CreateIndexMigrationAction(
                        index_name=index_name,
                        table_name=table_name,
                        columns=[constraint_column]
                    ), foreign_key)


def contains_index(inspector: Inspector, table: Union[sa.Table, str], column_name: str):
    table_name = table.name if isinstance(table, sa.Table) else table
    return any(index["column_names"] == [column_name] for index in inspector.get_indexes(table_name=table_name))


def handle_index_migration(
    table: sa.Table,
    table_name: str,
    old: sa.Column,
    new: sa.Column,
    column_name: str,
    handler: MigrationHandler,
    inspector: Inspector,
    foreign_key: bool,
    removed: list,
    renamed: bool
):
    index_name = get_pg_name(f'ix_{table_name}_{column_name}')

    constraint_column = old.name if renamed else column_name
    indexes = inspector.get_indexes(table_name=table.name)
    if (new.index or isinstance(new.type, geoalchemy2.types.Geometry)) and not renamed and not contains_index(inspector, table, constraint_column):
        handler.add_action(ma.CreateIndexMigrationAction(
            index_name=index_name,
            table_name=table_name,
            columns=[constraint_column],
            using="gist" if isinstance(new.type, geoalchemy2.types.Geometry) else None
        ), foreign_key)
    else:
        for index in indexes:
            dropped = False
            if index["column_names"] == [constraint_column] and index["name"] not in removed:
                index_name = get_pg_name(rename_index_name(index["name"], table.name, table_name, old.name, new.name))
                removed.append(index["name"])
                if renamed or (not new.index and not isinstance(new.type, geoalchemy2.types.Geometry)):
                    dropped = True
                    handler.add_action(
                        ma.DropIndexMigrationAction(
                            index_name=index["name"],
                            table_name=table_name
                        ), foreign_key)
                if (new.index or isinstance(new.type, geoalchemy2.types.Geometry)) and (dropped or not contains_index(inspector, table, constraint_column)):
                    handler.add_action(ma.CreateIndexMigrationAction(
                        index_name=index_name,
                        table_name=table_name,
                        columns=[column_name],
                        using="gist" if isinstance(new.type, geoalchemy2.types.Geometry) else None
                    ), foreign_key)



from typing import Iterator
from typing import overload

from spinta import commands
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.ufuncs.query.components import PgQueryBuilder
from spinta.components import Context
from spinta.components import Model
from spinta.core.ufuncs import Expr
from spinta.exceptions import ItemDoesNotExist
from spinta.exceptions import NotFoundError
from spinta.typing import ObjectData
from spinta.ufuncs.basequerybuilder.components import QueryParams
from spinta.ufuncs.basequerybuilder.helpers import get_page_values
from spinta.ufuncs.resultbuilder.helpers import get_row_value
from spinta.utils.nestedstruct import flat_dicts_to_nested


@overload
@commands.getone.register(Context, Model, PostgreSQL)
def getone(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    id_: str,
) -> ObjectData:
    connection = context.get('transaction').connection
    table = backend.get_table(model)
    try:
        result = backend.get(connection, table, table.c._id == id_)
    except NotFoundError:
        raise ItemDoesNotExist(model, id=id_)
    data = flat_dicts_to_nested(dict(result))
    data['_type'] = model.model_type()
    return commands.cast_backend_to_python(context, model, backend, data)


@commands.getall.register(Context, Model, PostgreSQL)
def getall(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    query: Expr = None,
    default_expand: bool = True,
    params: QueryParams = None,
    **kwargs
) -> Iterator[ObjectData]:
    assert isinstance(query, (Expr, type(None))), query
    connection = context.get('transaction').connection

    if default_expand:
        if params is None:
            params = QueryParams()
        params.default_expand = default_expand

    builder = PgQueryBuilder(context)
    builder.update(model=model)
    table = backend.get_table(model)
    env = builder.init(backend, table, params)
    expr = env.resolve(query)
    where = env.execute(expr)
    qry = env.build(where)

    conn = connection.execution_options(stream_results=True)
    result = conn.execute(qry)
    for row in result:
        res = {}
        list_keys = []
        for key, sel in env.selected.items():
            if key in model.flatprops:
                prop = model.flatprops[key]
                if prop.list is not None and prop.list.place not in list_keys:
                    list_keys.append(prop.list.place)
            val = get_row_value(context, backend, row, sel, False)
            res[key] = val
        if model.page.is_enabled:
            res['_page'] = get_page_values(env, row)

        res['_type'] = model.model_type()

        res = flat_dicts_to_nested(res, list_keys=list_keys)
        res = commands.cast_backend_to_python(context, model, backend, res)
        yield res
